import json
import os
from textwrap import dedent

import pytest

from caper.caper_workflow_opts import CaperWorkflowOpts
from caper.cromwell_backend import BACKEND_AWS, BACKEND_GCP


def test_create_file(tmp_path):
    """Test without docker/singularity.
    """
    use_google_cloud_life_sciences = False
    gcp_zones = ['us-west-1', 'us-west-2']
    slurm_partition = 'my_partition'
    slurm_account = 'my_account'
    slurm_extra_param = 'my_extra_param'
    sge_pe = 'my_pe'
    sge_queue = 'my_queue'
    sge_extra_param = 'my_extra_param'
    pbs_queue = 'my_queue'
    pbs_extra_param = 'my_extra_param'

    co = CaperWorkflowOpts(
        use_google_cloud_life_sciences=use_google_cloud_life_sciences,
        gcp_zones=gcp_zones,
        slurm_partition=slurm_partition,
        slurm_account=slurm_account,
        slurm_extra_param=slurm_extra_param,
        sge_pe=sge_pe,
        sge_queue=sge_queue,
        sge_extra_param=sge_extra_param,
        pbs_queue=pbs_queue,
        pbs_extra_param=pbs_extra_param,
    )

    wdl = tmp_path / 'test.wdl'
    wdl.write_text('')

    inputs = None

    # check if backend and slurm_partition is replaced with
    # that of this custom options file.
    custom_options = tmp_path / 'my_custom_options.json'
    custom_options_dict = {
        'backend': 'world',
        CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES: {
            'slurm_partition': 'not_my_partition'
        },
    }
    custom_options.write_text(json.dumps(custom_options_dict, indent=4))

    backend = 'my_backend'
    max_retries = 999
    memory_retry_multiplier = 1.3
    gcp_monitoring_script = 'gs://dummy/gcp_monitoring_script.sh'
    basename = 'my_basename.json'

    f = co.create_file(
        directory=str(tmp_path),
        wdl=str(wdl),
        inputs=inputs,
        custom_options=str(custom_options),
        docker=None,
        singularity=None,
        singularity_cachedir=None,
        no_build_singularity=False,
        backend=backend,
        max_retries=max_retries,
        memory_retry_multiplier=memory_retry_multiplier,
        gcp_monitoring_script=gcp_monitoring_script,
        basename=basename,
    )

    with open(f) as fp:
        d = json.loads(fp.read())

    dra = d[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]
    assert dra['zones'] == ' '.join(gcp_zones)
    assert dra['slurm_partition'] == 'not_my_partition'
    assert dra['slurm_account'] == slurm_account
    assert dra['slurm_extra_param'] == slurm_extra_param
    assert dra['sge_pe'] == sge_pe
    assert dra['sge_queue'] == sge_queue
    assert dra['sge_extra_param'] == sge_extra_param
    assert dra['pbs_queue'] == pbs_queue
    assert dra['pbs_extra_param'] == pbs_extra_param

    assert d['backend'] == 'world'
    assert dra['maxRetries'] == max_retries
    # assert d['memory_retry_multiplier'] == memory_retry_multiplier
    # this should be ignored for non-gcp backends
    assert 'monitoring_script' not in d
    assert os.path.basename(f) == basename
    assert os.path.dirname(f) == str(tmp_path)

    # test for gcp backend
    f = co.create_file(
        directory=str(tmp_path),
        wdl=str(wdl),
        backend='gcp',
        docker='ubuntu:latest',
        max_retries=max_retries,
        gcp_monitoring_script=gcp_monitoring_script,
        basename=basename,
    )
    with open(f) as fp:
        d = json.loads(fp.read())
    assert d['monitoring_script'] == gcp_monitoring_script


def test_create_file_with_google_cloud_life_sciences(tmp_path):
    """Test with use_google_cloud_life_sciences flag.
    zones should not be written to dra.
    """
    gcp_zones = ['us-west-1', 'us-west-2']

    co = CaperWorkflowOpts(use_google_cloud_life_sciences=True, gcp_zones=gcp_zones)

    wdl = tmp_path / 'test.wdl'
    wdl.write_text('')

    f = co.create_file(directory=str(tmp_path), wdl=str(wdl))

    with open(f) as fp:
        d = json.loads(fp.read())

    dra = d[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]
    assert 'zones' not in dra


def test_create_file_docker(tmp_path):
    """Test with docker and docker defined in WDL.
    """
    wdl_contents = dedent(
        """\
        version 1.0
        workflow test_docker {
            meta {
                caper_docker: "ubuntu:latest"
            }
        }
    """
    )

    wdl = tmp_path / 'docker.wdl'
    wdl.write_text(wdl_contents)

    co = CaperWorkflowOpts()

    # cloud backend gcp should try to find docker in WDL
    f_gcp = co.create_file(
        directory=str(tmp_path),
        wdl=str(wdl),
        backend=BACKEND_GCP,
        basename='opts_gcp.json',
    )
    with open(f_gcp) as fp:
        d_gcp = json.loads(fp.read())
        dra_gcp = d_gcp[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]
    assert dra_gcp['docker'] == 'ubuntu:latest'

    # cloud backend aws should try to find docker in WDL
    f_aws = co.create_file(
        directory=str(tmp_path),
        wdl=str(wdl),
        backend=BACKEND_AWS,
        basename='opts_aws.json',
    )
    with open(f_aws) as fp:
        d_aws = json.loads(fp.read())
        dra_aws = d_aws[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]
    assert dra_aws['docker'] == 'ubuntu:latest'

    # local backend should not try to find docker in WDL
    # if docker is not defined
    f_local = co.create_file(
        directory=str(tmp_path),
        wdl=str(wdl),
        backend='my_backend',
        basename='opts_local.json',
    )
    with open(f_local) as fp:
        d_local = json.loads(fp.read())
        dra_local = d_local[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]
    assert 'docker' not in dra_local

    # local backend should use docker if docker is explicitly defined
    f_local2 = co.create_file(
        directory=str(tmp_path),
        wdl=str(wdl),
        docker='ubuntu:16',
        backend='my_backend',
        basename='opts_local2.json',
    )
    with open(f_local2) as fp:
        d_local2 = json.loads(fp.read())
        dra_local2 = d_local2[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]
    assert dra_local2['docker'] == 'ubuntu:16'


def test_create_file_singularity(tmp_path):
    """Test with singularity and singularity defined in WDL.
    """
    wdl_contents = dedent(
        """\
        version 1.0
        workflow test_singularity {
            meta {
                caper_docker: "ubuntu:latest"
                caper_singularity: "docker://ubuntu:latest"
            }
        }
    """
    )

    wdl = tmp_path / 'singularity.wdl'
    wdl.write_text(wdl_contents)

    co = CaperWorkflowOpts()

    # cloud backend gcp should not try to find singularity in WDL
    f_gcp = co.create_file(
        directory=str(tmp_path),
        wdl=str(wdl),
        backend=BACKEND_GCP,
        basename='opts_gcp.json',
    )
    with open(f_gcp) as fp:
        d_gcp = json.loads(fp.read())
        dra_gcp = d_gcp[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]
    assert 'singularity' not in dra_gcp

    # cloud backend aws should not try to find singularity in WDL
    f_aws = co.create_file(
        directory=str(tmp_path),
        wdl=str(wdl),
        backend=BACKEND_AWS,
        basename='opts_aws.json',
    )
    with open(f_aws) as fp:
        d_aws = json.loads(fp.read())
        dra_aws = d_aws[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]
    assert 'singularity' not in dra_aws

    # cloud backend aws/gcp should not work with singularity
    with pytest.raises(ValueError):
        co.create_file(
            directory=str(tmp_path),
            wdl=str(wdl),
            backend=BACKEND_GCP,
            singularity='',
            basename='opts_gcp2.json',
        )
    with pytest.raises(ValueError):
        co.create_file(
            directory=str(tmp_path),
            wdl=str(wdl),
            backend=BACKEND_AWS,
            singularity='',
            basename='opts_aws2.json',
        )

    # local backend should not try to find singularity in WDL
    # if singularity is not defined
    f_local = co.create_file(
        directory=str(tmp_path),
        wdl=str(wdl),
        backend='my_backend',
        basename='opts_local.json',
    )
    with open(f_local) as fp:
        d_local = json.loads(fp.read())
        dra_local = d_local[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]
    assert 'singularity' not in dra_local

    # input JSON to test singularity bindpath
    # this will be test thoroughly in other testing module (test_singularity)
    inputs = tmp_path / 'inputs.json'
    inputs_dict = {
        'test.input': '/a/b/c/d.txt',
        'test.input2': '/a/b/e.txt',
        'test.input3': '/f/g/h.txt',
    }
    inputs.write_text(json.dumps(inputs_dict, indent=4))

    # local backend should use singularity if singularity is explicitly defined
    # also, singularity_bindpath should be input JSON.
    f_local2 = co.create_file(
        directory=str(tmp_path),
        wdl=str(wdl),
        inputs=str(inputs),
        singularity='ubuntu:16',
        singularity_cachedir='/tmp',
        no_build_singularity=True,
        backend='my_backend',
        basename='opts_local2.json',
    )
    with open(f_local2) as fp:
        d_local2 = json.loads(fp.read())
        dra_local2 = d_local2[CaperWorkflowOpts.DEFAULT_RUNTIME_ATTRIBUTES]
    assert dra_local2['singularity'] == 'ubuntu:16'
    assert dra_local2['singularity_cachedir'] == '/tmp'
    assert sorted(dra_local2['singularity_bindpath'].split(',')) == ['/a/b', '/f/g']



import json
import logging
from collections import UserDict
from copy import deepcopy
from textwrap import dedent

from .dict_tool import merge_dict

logger = logging.getLogger(__name__)


BACKEND_GCP = 'gcp'
BACKEND_AWS = 'aws'
BACKEND_LOCAL = 'Local'
BACKEND_ALIAS_LOCAL = 'local'
BACKEND_SLURM = 'slurm'
BACKEND_SGE = 'sge'
BACKEND_PBS = 'pbs'
DEFAULT_BACKEND = BACKEND_LOCAL

FILESYSTEM_GCS = 'gcs'
FILESYSTEM_LOCAL = 'local'
FILESYSTEM_S3 = 's3'

LOCALIZATION_STRAT_COPY = 'copy'
LOCALIZATION_STRAT_HARDLINK = 'hard-link'
LOCALIZATION_STRAT_SOFTLINK = 'soft-link'

CALL_CACHING_DUP_STRAT_COPY = 'copy'
CALL_CACHING_DUP_STRAT_HARDLINK = 'hard-link'
CALL_CACHING_DUP_STRAT_REFERENCE = 'reference'
CALL_CACHING_DUP_STRAT_SOFTLINK = 'soft-link'

LOCAL_HASH_STRAT_FILE = 'file'
LOCAL_HASH_STRAT_PATH = 'path'
LOCAL_HASH_STRAT_PATH_MTIME = 'path+modtime'
SOFT_GLOB_OUTPUT_CMD = 'ln -sL GLOB_PATTERN GLOB_DIRECTORY 2> /dev/null'


def get_s3_bucket_name(s3_uri):
    return s3_uri.replace('s3://', '', 1).split('/')[0]


class CromwellBackendCommon(UserDict):
    """Basic stanzas for Cromwell backend conf.
    """

    TEMPLATE = {
        'backend': {},
        'webservice': {},
        'services': {
            'LoadController': {
                'class': 'cromwell.services.loadcontroller.impl'
                '.LoadControllerServiceActor',
                'config': {
                    # added due to issues on stanford sherlock/scg
                    'control-frequency': '21474834 seconds'
                },
            }
        },
        'system': {
            'job-rate-control': {'jobs': 1, 'per': '2 seconds'},
            'abort-jobs-on-terminate': True,
            'graceful-server-shutdown': True,
        },
        'call-caching': {'invalidate-bad-cache-results': True},
        'akka': {'http': {'server': {'request-timeout': '60 seconds'}}},
    }

    DEFAULT_MAX_CONCURRENT_WORKFLOWS = 40
    DEFAULT_MEMORY_RETRY_ERROR_KEYS = ('OutOfMemory', 'Killed')

    def __init__(
        self,
        default_backend,
        disable_call_caching=False,
        max_concurrent_workflows=DEFAULT_MAX_CONCURRENT_WORKFLOWS,
        memory_retry_error_keys=DEFAULT_MEMORY_RETRY_ERROR_KEYS,
    ):
        """
        Args:
            memory_retry_error_keys:
                List of error strings to catch out-of-memory error
                See https://cromwell.readthedocs.io/en/develop/cromwell_features/RetryWithMoreMemory
                for details.
        """
        super().__init__(deepcopy(CromwellBackendCommon.TEMPLATE))

        if default_backend is None:
            default_backend = DEFAULT_BACKEND
        self['backend']['default'] = default_backend
        self['call-caching']['enabled'] = not disable_call_caching
        self['system']['max-concurrent-workflows'] = max_concurrent_workflows
        # Cromwell's bug in memory-retry feature.
        # Disabled until it's fixed on Cromwell's side.
        # if memory_retry_error_keys:
        #     if isinstance(memory_retry_error_keys, tuple):
        #         memory_retry_error_keys = list(memory_retry_error_keys)
        #     self['system']['memory-retry-error-keys'] = memory_retry_error_keys


class CromwellBackendServer(UserDict):
    """Stanzas for Cromwell server.
    """

    TEMPLATE = {'webservice': {}}

    DEFAULT_SERVER_PORT = 8000

    def __init__(self, server_port=DEFAULT_SERVER_PORT):
        super().__init__(deepcopy(CromwellBackendServer.TEMPLATE))

        self['webservice']['port'] = server_port


class CromwellBackendDatabase(UserDict):
    """Common stanzas for Cromwell's metadata database.
    """

    TEMPLATE = {'database': {'db': {'connectionTimeout': 5000, 'numThreads': 1}}}

    DB_IN_MEMORY = 'in-memory'
    DB_FILE = 'file'
    DB_MYSQL = 'mysql'
    DB_POSTGRESQL = 'postgresql'

    PROFILE_MYSQL = 'slick.jdbc.MySQLProfile$'
    PROFILE_POSTGRESQL = 'slick.jdbc.PostgresProfile$'
    JDBC_DRIVER_MYSQL = 'com.mysql.cj.jdbc.Driver'
    JDBC_DRIVER_POSTGRESQL = 'org.postgresql.Driver'
    JDBC_URL_FILE = (
        'jdbc:hsqldb:file:{file};shutdown=false;hsqldb.tx=mvcc;'
        'hsqldb.lob_compressed=true;'
        'hsqldb.default_table_type=cached;'
        'hsqldb.result_max_memory_rows=10000;'
        'hsqldb.large_data=true;'
        'hsqldb.applog=1;'
        'hsqldb.script_format=3'
    )
    JDBC_URL_MYSQL = (
        'jdbc:mysql://{ip}:{port}/{name}?'
        'allowPublicKeyRetrieval=true&useSSL=false&'
        'rewriteBatchedStatements=true&serverTimezone=UTC'
    )
    JDBC_URL_POSTGRESQL = 'jdbc:postgresql://{ip}:{port}/{name}'

    DEFAULT_DB = DB_IN_MEMORY
    DEFAULT_DB_TIMEOUT_MS = 30000
    DEFAULT_MYSQL_DB_IP = 'localhost'
    DEFAULT_MYSQL_DB_PORT = 3306
    DEFAULT_MYSQL_DB_USER = 'cromwell'
    DEFAULT_MYSQL_DB_PASSWORD = 'cromwell'
    DEFAULT_MYSQL_DB_NAME = 'cromwell'
    DEFAULT_POSTGRESQL_DB_IP = 'localhost'
    DEFAULT_POSTGRESQL_DB_PORT = 5432
    DEFAULT_POSTGRESQL_DB_USER = 'cromwell'
    DEFAULT_POSTGRESQL_DB_PASSWORD = 'cromwell'
    DEFAULT_POSTGRESQL_DB_NAME = 'cromwell'

    def __init__(
        self,
        db=DEFAULT_DB,
        db_timeout=DEFAULT_DB_TIMEOUT_MS,
        mysql_db_ip=DEFAULT_MYSQL_DB_IP,
        mysql_db_port=DEFAULT_MYSQL_DB_PORT,
        mysql_db_user=DEFAULT_MYSQL_DB_USER,
        mysql_db_password=DEFAULT_MYSQL_DB_PASSWORD,
        mysql_db_name=DEFAULT_MYSQL_DB_NAME,
        postgresql_db_ip=DEFAULT_POSTGRESQL_DB_IP,
        postgresql_db_port=DEFAULT_POSTGRESQL_DB_PORT,
        postgresql_db_user=DEFAULT_POSTGRESQL_DB_USER,
        postgresql_db_password=DEFAULT_POSTGRESQL_DB_PASSWORD,
        postgresql_db_name=DEFAULT_POSTGRESQL_DB_NAME,
        file_db=None,
    ):
        super().__init__(deepcopy(CromwellBackendDatabase.TEMPLATE))

        database = self['database']
        db_obj = database['db']

        db_obj['connectionTimeout'] = db_timeout

        if db == CromwellBackendDatabase.DB_FILE:
            if not file_db:
                raise ValueError('file_db must be defined for db {db}'.format(db=db))

        if db == CromwellBackendDatabase.DB_IN_MEMORY:
            pass

        elif db == CromwellBackendDatabase.DB_FILE:
            db_obj['url'] = CromwellBackendDatabase.JDBC_URL_FILE.format(file=file_db)

        elif db == CromwellBackendDatabase.DB_MYSQL:
            database['profile'] = CromwellBackendDatabase.PROFILE_MYSQL
            db_obj['driver'] = CromwellBackendDatabase.JDBC_DRIVER_MYSQL
            db_obj['url'] = CromwellBackendDatabase.JDBC_URL_MYSQL.format(
                ip=mysql_db_ip, port=mysql_db_port, name=mysql_db_name
            )
            db_obj['user'] = mysql_db_user
            db_obj['password'] = mysql_db_password

        elif db == CromwellBackendDatabase.DB_POSTGRESQL:
            database['profile'] = CromwellBackendDatabase.PROFILE_POSTGRESQL
            db_obj['driver'] = CromwellBackendDatabase.JDBC_DRIVER_POSTGRESQL
            db_obj['url'] = CromwellBackendDatabase.JDBC_URL_POSTGRESQL.format(
                ip=postgresql_db_ip, port=postgresql_db_port, name=postgresql_db_name
            )
            db_obj['port'] = postgresql_db_port
            db_obj['user'] = postgresql_db_user
            db_obj['password'] = postgresql_db_password

        else:
            raise ValueError('Unsupported DB type {db}'.format(db=db))


class CromwellBackendBase(UserDict):
    """Base skeleton backend for all backends
    """

    TEMPLATE = {'backend': {'providers': {}}}
    TEMPLATE_BACKEND = {'config': {'default-runtime-attributes': {}, 'filesystems': {}}}
    DEFAULT_CALL_CACHING_DUP_STRAT = (
        CALL_CACHING_DUP_STRAT_SOFTLINK,
        CALL_CACHING_DUP_STRAT_HARDLINK,
        CALL_CACHING_DUP_STRAT_COPY,
    )
    DEFAULT_CONCURRENT_JOB_LIMIT = 1000

    def __init__(
        self,
        backend_name,
        max_concurrent_tasks=DEFAULT_CONCURRENT_JOB_LIMIT,
        filesystem_name=None,
        call_caching_dup_strat=DEFAULT_CALL_CACHING_DUP_STRAT,
    ):
        """
        Args:
            backend_name:
                Backend's name.
            max_concurrent_tasks:
                Maximum number of tasks (regardless of number of workflows).
            filesystem_name:
                Filesystem's name to set up call-caching strategy within.
            call_caching_dup_strat:
                Call-caching strategy string.
                This can be either a strategy or a list of strategies.
        """
        super().__init__(deepcopy(CromwellBackendBase.TEMPLATE))

        if backend_name is None:
            raise ValueError('backend_name must be provided.')
        self._backend_name = backend_name

        self.backend = CromwellBackendBase.TEMPLATE_BACKEND

        config = self.backend_config
        config['concurrent-job-limit'] = max_concurrent_tasks

        if filesystem_name:
            if isinstance(call_caching_dup_strat, tuple):
                call_caching_dup_strat = list(call_caching_dup_strat)

            config['filesystems'][filesystem_name] = {
                'caching': {'duplication-strategy': call_caching_dup_strat}
            }

    @property
    def backend(self):
        return self['backend']['providers'][self._backend_name]

    @backend.setter
    def backend(self, backend):
        self['backend']['providers'][self._backend_name] = deepcopy(backend)

    def merge_backend(self, backend):
        merge_dict(self.backend, backend)

    @property
    def backend_config(self):
        return self.backend['config']

    @property
    def default_runtime_attributes(self):
        """Backend's default runtime attributes in self.backend_config.
        """
        return self.backend_config['default-runtime-attributes']


class CromwellBackendGCP(CromwellBackendBase):
    TEMPLATE = {'google': {'application-name': 'cromwell'}}
    TEMPLATE_BACKEND = {
        'config': {
            'default-runtime-attributes': {},
            'genomics-api-queries-per-100-seconds': 1000,
            'maximum-polling-interval': 600,
            'localization-attempts': 3,
            'genomics': {
                'restrict-metadata-access': False,
                'compute-service-account': 'default',
            },
        }
    }
    ACTOR_FACTORY_V2ALPHA = (
        'cromwell.backend.google.pipelines.v2alpha1.PipelinesApiLifecycleActorFactory'
    )
    ACTOR_FACTORY_V2BETA = (
        'cromwell.backend.google.pipelines.v2beta.PipelinesApiLifecycleActorFactory'
    )
    GENOMICS_ENDPOINT_V2ALPHA = 'https://genomics.googleapis.com/'
    GENOMICS_ENDPOINT_V2BETA = 'https://lifesciences.googleapis.com/'
    DEFAULT_REGION = 'us-central1'
    DEFAULT_CALL_CACHING_DUP_STRAT = CALL_CACHING_DUP_STRAT_REFERENCE

    def __init__(
        self,
        gcp_prj,
        gcp_out_dir,
        gcp_service_account_key_json=None,
        use_google_cloud_life_sciences=False,
        gcp_region=DEFAULT_REGION,
        gcp_zones=None,
        max_concurrent_tasks=CromwellBackendBase.DEFAULT_CONCURRENT_JOB_LIMIT,
        call_caching_dup_strat=DEFAULT_CALL_CACHING_DUP_STRAT,
    ):
        """
        Args:
            gcp_service_account_key_json:
                Use this key JSON file to use service_account scheme
                instead of application_default.
            use_google_cloud_life_sciences:
                Use Google Cloud Life Sciences API (v2beta) instead of
                deprecated Genomics API (v2alpha1).
            gcp_region:
                Region for Google Cloud Life Sciences API.
            gcp_zones:
                List of zones for Genomics API.
                Ignored if use_google_cloud_life_sciences.
        """
        if call_caching_dup_strat not in (
            CALL_CACHING_DUP_STRAT_REFERENCE,
            CALL_CACHING_DUP_STRAT_COPY,
        ):
            raise ValueError(
                'Wrong call_caching_dup_strat for GCP: {v}'.format(
                    v=call_caching_dup_strat
                )
            )

        super().__init__(
            backend_name=BACKEND_GCP,
            max_concurrent_tasks=max_concurrent_tasks,
            filesystem_name=FILESYSTEM_GCS,
            call_caching_dup_strat=call_caching_dup_strat,
        )
        merge_dict(self.data, CromwellBackendGCP.TEMPLATE)
        self.merge_backend(CromwellBackendGCP.TEMPLATE_BACKEND)

        config = self.backend_config
        genomics = config['genomics']
        filesystems = config['filesystems']

        if gcp_service_account_key_json:
            genomics['auth'] = 'service-account'
            filesystems[FILESYSTEM_GCS]['auth'] = 'service-account'
            self['google']['auths'] = [
                {
                    'name': 'service-account',
                    'scheme': 'service_account',
                    'json-file': gcp_service_account_key_json,
                }
            ]
            # parse service account key JSON to get client_email.
            with open(gcp_service_account_key_json) as fp:
                key_json = json.loads(fp.read())
            genomics['compute-service-account'] = key_json['client_email']
        else:
            genomics['auth'] = 'application-default'
            filesystems[FILESYSTEM_GCS]['auth'] = 'application-default'
            self['google']['auths'] = [
                {'name': 'application-default', 'scheme': 'application_default'}
            ]

        if use_google_cloud_life_sciences:
            self.backend['actor-factory'] = CromwellBackendGCP.ACTOR_FACTORY_V2BETA
            genomics['endpoint-url'] = CromwellBackendGCP.GENOMICS_ENDPOINT_V2BETA
            genomics['location'] = gcp_region
        else:
            self.backend['actor-factory'] = CromwellBackendGCP.ACTOR_FACTORY_V2ALPHA
            genomics['endpoint-url'] = CromwellBackendGCP.GENOMICS_ENDPOINT_V2ALPHA
            if gcp_zones:
                self.default_runtime_attributes['zones'] = ' '.join(gcp_zones)

        config['project'] = gcp_prj

        if not gcp_out_dir.startswith('gs://'):
            raise ValueError(
                'Wrong GCS bucket URI for gcp_out_dir: {v}'.format(v=gcp_out_dir)
            )
        config['root'] = gcp_out_dir


class CromwellBackendAWS(CromwellBackendBase):
    TEMPLATE = {
        'aws': {
            'application-name': 'cromwell',
            'auths': [{'name': 'default', 'scheme': 'default'}],
        },
        'engine': {'filesystems': {FILESYSTEM_S3: {'auth': 'default'}}},
    }
    TEMPLATE_BACKEND = {
        'actor-factory': 'cromwell.backend.impl.aws.AwsBatchBackendLifecycleActorFactory',
        'config': {
            'default-runtime-attributes': {},
            'numSubmitAttempts': 6,
            'numCreateDefinitionAttempts': 6,
            'auth': 'default',
            'filesystems': {FILESYSTEM_S3: {'auth': 'default'}},
        },
    }
    DEFAULT_CALL_CACHING_DUP_STRAT = CALL_CACHING_DUP_STRAT_REFERENCE

    def __init__(
        self,
        aws_batch_arn,
        aws_region,
        aws_out_dir,
        max_concurrent_tasks=CromwellBackendBase.DEFAULT_CONCURRENT_JOB_LIMIT,
        call_caching_dup_strat=DEFAULT_CALL_CACHING_DUP_STRAT,
    ):
        if call_caching_dup_strat not in (
            CALL_CACHING_DUP_STRAT_REFERENCE,
            CALL_CACHING_DUP_STRAT_COPY,
        ):
            raise ValueError(
                'Wrong call_caching_dup_strat for S3: {v}'.format(
                    v=call_caching_dup_strat
                )
            )
        if call_caching_dup_strat == CALL_CACHING_DUP_STRAT_REFERENCE:
            logger.warning(
                'Warning for aws backend: "reference" mode for call_caching_dup_strat currently '
                'does not work with Cromwell<=61. Cromwell will still use the "copy" mode '
                'It will make cache copies for all call-cached outputs, which will lead to '
                'unnecessary extra charge for the output S3 bucket. '
                '"reference" mode on gcp backend works fine. '
                'See the following link for details. '
                'https://github.com/broadinstitute/cromwell/issues/6327. '
                'It is recommend to clean up previous workflow\'s outputs manually '
                'with "caper cleanup WORKFLOW_ID_OR_METADATA_JSON_FILE" or '
                'with AWS CLI. e.g. '
                '"aws s3 rm --recursive s3://some-bucket/a/b/c/WORKFLOW_ID". '
            )
        super().__init__(
            backend_name=BACKEND_AWS,
            max_concurrent_tasks=max_concurrent_tasks,
            filesystem_name=FILESYSTEM_S3,
            call_caching_dup_strat=call_caching_dup_strat,
        )
        merge_dict(self.data, CromwellBackendAWS.TEMPLATE)
        self.merge_backend(CromwellBackendAWS.TEMPLATE_BACKEND)

        aws = self[BACKEND_AWS]
        aws['region'] = aws_region

        config = self.backend_config
        if not aws_out_dir.startswith('s3://'):
            raise ValueError(
                'Wrong S3 bucket URI for aws_out_dir: {v}'.format(v=aws_out_dir)
            )
        config['root'] = aws_out_dir
        self.default_runtime_attributes['scriptBucketName'] = get_s3_bucket_name(
            aws_out_dir
        )
        self.default_runtime_attributes['queueArn'] = aws_batch_arn


class CromwellBackendLocal(CromwellBackendBase):
    """Class constants:
        MAKE_CMD_SUBMIT:
            Includes BASH command line for Singularity.
    """

    TEMPLATE_BACKEND = {
        'actor-factory': 'cromwell.backend.impl.sfs.config.ConfigBackendLifecycleActorFactory',
        'config': {
            'script-epilogue': 'sleep 5',
            'filesystems': {
                FILESYSTEM_LOCAL: {
                    'localization': [
                        LOCALIZATION_STRAT_SOFTLINK,
                        LOCALIZATION_STRAT_HARDLINK,
                        LOCALIZATION_STRAT_COPY,
                    ],
                    'caching': {'check-sibling-md5': True},
                }
            },
            'run-in-background': True,
            'runtime-attributes': dedent(
                """\
                Int? gpu
                String? docker
                String? docker_user
                String? singularity
                String? singularity_bindpath
                String? singularity_cachedir
                """
            ),
            'submit': dedent(
                """\
                if [ -z \\"$SINGULARITY_BINDPATH\\" ]; then export SINGULARITY_BINDPATH=${singularity_bindpath}; fi; \\
                if [ -z \\"$SINGULARITY_CACHEDIR\\" ]; then export SINGULARITY_CACHEDIR=${singularity_cachedir}; fi;
                ${if !defined(singularity) then '/bin/bash ' + script
                  else
                    'singularity exec --cleanenv ' +
                    '--home ' + cwd + ' ' +
                    (if defined(gpu) then '--nv ' else '') +
                    singularity + ' /bin/bash ' + script}
            """
            ),
            'submit-docker': dedent(
                """\
                rm -f ${docker_cid}
                docker run \\
                  --cidfile ${docker_cid} \\
                  -i \\
                  ${'--user ' + docker_user} \\
                  --entrypoint ${job_shell} \\
                  -v ${cwd}:${docker_cwd} \\
                  ${docker} ${docker_script}
                """
            ),
        },
    }

    DEFAULT_LOCAL_HASH_STRAT = LOCAL_HASH_STRAT_PATH_MTIME

    def __init__(
        self,
        local_out_dir,
        backend_name=BACKEND_LOCAL,
        soft_glob_output=False,
        local_hash_strat=DEFAULT_LOCAL_HASH_STRAT,
        max_concurrent_tasks=CromwellBackendBase.DEFAULT_CONCURRENT_JOB_LIMIT,
    ):
        super().__init__(
            backend_name=backend_name,
            max_concurrent_tasks=max_concurrent_tasks,
            filesystem_name=FILESYSTEM_LOCAL,
        )
        self.merge_backend(CromwellBackendLocal.TEMPLATE_BACKEND)

        config = self.backend_config
        filesystem_local = config['filesystems'][FILESYSTEM_LOCAL]
        caching = filesystem_local['caching']

        if local_hash_strat not in (
            LOCAL_HASH_STRAT_FILE,
            LOCAL_HASH_STRAT_PATH,
            LOCAL_HASH_STRAT_PATH_MTIME,
        ):
            raise ValueError(
                'Wrong local_hash_strat: {strat}'.format(strat=local_hash_strat)
            )
        caching['hashing-strategy'] = local_hash_strat

        if soft_glob_output:
            config['glob-link-command'] = SOFT_GLOB_OUTPUT_CMD

        if local_out_dir is None:
            raise ValueError('local_out_dir must be provided.')
        config['root'] = local_out_dir


class CromwellBackendSLURM(CromwellBackendLocal):
    """SLURM backend.
    Try sbatching up to 3 times every 30 second.
    Some busy SLURM clusters spit out error, which results in a failure of the whole workflow

    Squeues every 30 second (up to 3 times)
    Unlike qstat -j JOB_ID, squeue -j JOB_ID doesn't return 1 when there is no such job
    So we need to use squeue -j JOB_ID --noheader and check if output is empty
    Try polling up to 3 times since squeue fails on some busy SLURM clusters
    e.g. on Stanford Sherlock, squeue didn't work when server is busy
    """

    TEMPLATE_BACKEND = {
        'config': {
            'default-runtime-attributes': {'time': 24},
            'exit-code-timeout-seconds': 360,
            'runtime-attributes': dedent(
                """\
                String? docker
                String? docker_user
                Int cpu = 1
                Int? gpu
                Int? time
                Int? memory_mb
                String? slurm_partition
                String? slurm_account
                String? slurm_extra_param
                String? singularity
                String? singularity_bindpath
                String? singularity_cachedir
            """
            ),
            'submit': dedent(
                """\
                if [ -z \\"$SINGULARITY_BINDPATH\\" ]; then export SINGULARITY_BINDPATH=${singularity_bindpath}; fi; \\
                if [ -z \\"$SINGULARITY_CACHEDIR\\" ]; then export SINGULARITY_CACHEDIR=${singularity_cachedir}; fi;

                ITER=0
                until [ $ITER -ge 3 ]; do
                    sbatch \\
                        --export=ALL \\
                        -J ${job_name} \\
                        -D ${cwd} \\
                        -o ${out} \\
                        -e ${err} \\
                        ${'-t ' + time*60} \\
                        -n 1 \\
                        --ntasks-per-node=1 \\
                        ${'--cpus-per-task=' + cpu} \\
                        ${true="--mem=" false="" defined(memory_mb)}${memory_mb} \\
                        ${'-p ' + slurm_partition} \\
                        ${'--account ' + slurm_account} \\
                        ${'--gres gpu:' + gpu} \\
                        ${slurm_extra_param} \\
                        --wrap "${if !defined(singularity) then '/bin/bash ' + script
                                  else
                                    'singularity exec --cleanenv ' +
                                    '--home ' + cwd + ' ' +
                                    (if defined(gpu) then '--nv ' else '') +
                                    singularity + ' /bin/bash ' + script}" \\
                        && break
                    ITER=$[$ITER+1]
                    sleep 30
                done
            """
            ),
            'check-alive': dedent(
                """\
                for ITER in 1 2 3; do
                    CHK_ALIVE=$(squeue --noheader -j ${job_id} --format=%i | grep ${job_id})
                    if [ -z "$CHK_ALIVE" ]; then if [ "$ITER" == 3 ]; then /bin/bash -c 'exit 1'; else sleep 30; fi; else echo $CHK_ALIVE; break; fi
                done
            """
            ),
            'kill': 'scancel ${job_id}',
            'job-id-regex': 'Submitted batch job (\\d+).*',
        }
    }

    def __init__(
        self,
        local_out_dir,
        max_concurrent_tasks=CromwellBackendBase.DEFAULT_CONCURRENT_JOB_LIMIT,
        soft_glob_output=False,
        local_hash_strat=CromwellBackendLocal.DEFAULT_LOCAL_HASH_STRAT,
        slurm_partition=None,
        slurm_account=None,
        slurm_extra_param=None,
    ):
        super().__init__(
            local_out_dir=local_out_dir,
            backend_name=BACKEND_SLURM,
            max_concurrent_tasks=max_concurrent_tasks,
            soft_glob_output=soft_glob_output,
            local_hash_strat=local_hash_strat,
        )
        self.merge_backend(CromwellBackendSLURM.TEMPLATE_BACKEND)
        self.backend_config.pop('submit-docker')

        if slurm_partition:
            self.default_runtime_attributes['slurm_partition'] = slurm_partition
        if slurm_account:
            self.default_runtime_attributes['slurm_account'] = slurm_account
        if slurm_extra_param:
            self.default_runtime_attributes['slurm_extra_param'] = slurm_extra_param


class CromwellBackendSGE(CromwellBackendLocal):
    TEMPLATE_BACKEND = {
        'config': {
            'default-runtime-attributes': {'time': 24},
            'exit-code-timeout-seconds': 180,
            'runtime-attributes': dedent(
                """\
                String? docker
                String? docker_user
                String sge_pe = "shm"
                Int cpu = 1
                Int? gpu
                Int? time
                Int? memory_mb
                String? sge_queue
                String? sge_extra_param
                String? singularity
                String? singularity_bindpath
                String? singularity_cachedir
            """
            ),
            'submit': dedent(
                """\
                if [ -z \\"$SINGULARITY_BINDPATH\\" ]; then export SINGULARITY_BINDPATH=${singularity_bindpath}; fi; \\
                if [ -z \\"$SINGULARITY_CACHEDIR\\" ]; then export SINGULARITY_CACHEDIR=${singularity_cachedir}; fi;

                echo "${if !defined(singularity) then '/bin/bash ' + script
                        else
                          'singularity exec --cleanenv ' +
                          '--home ' + cwd + ' ' +
                          (if defined(gpu) then '--nv ' else '') +
                          singularity + ' /bin/bash ' + script}" | \\
                qsub \\
                    -S /bin/sh \\
                    -terse \\
                    -b n \\
                    -N ${job_name} \\
                    -wd ${cwd} \\
                    -o ${out} \\
                    -e ${err} \\
                    ${if cpu>1 then '-pe ' + sge_pe + ' ' else ''} \\
                    ${if cpu>1 then cpu else ''} \\
                    ${true="-l h_vmem=$(expr " false="" defined(memory_mb)}${memory_mb}${true=" / " false="" defined(memory_mb)}${if defined(memory_mb) then cpu else ""}${true=")m" false="" defined(memory_mb)} \\
                    ${true="-l s_vmem=$(expr " false="" defined(memory_mb)}${memory_mb}${true=" / " false="" defined(memory_mb)}${if defined(memory_mb) then cpu else ""}${true=")m" false="" defined(memory_mb)} \\
                    ${'-l h_rt=' + time + ':00:00'} \\
                    ${'-l s_rt=' + time + ':00:00'} \\
                    ${'-q ' + sge_queue} \\
                    ${'-l gpu=' + gpu} \\
                    ${sge_extra_param} \\
                    -V
            """
            ),
            'check-alive': 'qstat -j ${job_id}',
            'kill': 'qdel ${job_id}',
            'job-id-regex': '(\\d+)',
        }
    }

    def __init__(
        self,
        local_out_dir,
        max_concurrent_tasks=CromwellBackendBase.DEFAULT_CONCURRENT_JOB_LIMIT,
        soft_glob_output=False,
        local_hash_strat=CromwellBackendLocal.DEFAULT_LOCAL_HASH_STRAT,
        sge_pe=None,
        sge_queue=None,
        sge_extra_param=None,
    ):
        super().__init__(
            local_out_dir=local_out_dir,
            backend_name=BACKEND_SGE,
            max_concurrent_tasks=max_concurrent_tasks,
            soft_glob_output=soft_glob_output,
            local_hash_strat=local_hash_strat,
        )
        self.merge_backend(CromwellBackendSGE.TEMPLATE_BACKEND)
        self.backend_config.pop('submit-docker')

        if sge_pe:
            self.default_runtime_attributes['sge_pe'] = sge_pe
        if sge_queue:
            self.default_runtime_attributes['sge_queue'] = sge_queue
        if sge_extra_param:
            self.default_runtime_attributes['sge_extra_param'] = sge_extra_param


class CromwellBackendPBS(CromwellBackendLocal):
    TEMPLATE_BACKEND = {
        'config': {
            'default-runtime-attributes': {'time': 24},
            'script-epilogue': 'sleep 5',
            'runtime-attributes': dedent(
                """\
                String? docker
                String? docker_user
                Int cpu = 1
                Int? gpu
                Int? time
                Int? memory_mb
                String? pbs_queue
                String? pbs_extra_param
                String? singularity
                String? singularity_bindpath
                String? singularity_cachedir
            """
            ),
            'submit': dedent(
                """\
                if [ -z \\"$SINGULARITY_BINDPATH\\" ]; then export SINGULARITY_BINDPATH=${singularity_bindpath}; fi; \\
                if [ -z \\"$SINGULARITY_CACHEDIR\\" ]; then export SINGULARITY_CACHEDIR=${singularity_cachedir}; fi;

                echo "${if !defined(singularity) then '/bin/bash ' + script
                        else
                          'singularity exec --cleanenv ' +
                          '--home ' + cwd + ' ' +
                          (if defined(gpu) then '--nv ' else '') +
                          singularity + ' /bin/bash ' + script}" | \\
                qsub \\
                    -N ${job_name} \\
                    -o ${out} \\
                    -e ${err} \\
                    ${true="-lnodes=1:ppn=" false="" defined(cpu)}${cpu}${true=":mem=" false="" defined(memory_mb)}${memory_mb}${true="mb" false="" defined(memory_mb)} \\
                    ${'-lwalltime=' + time + ':0:0'} \\
                    ${'-lngpus=' + gpu} \\
                    ${'-q ' + pbs_queue} \\
                    ${pbs_extra_param} \\
                    -V
            """
            ),
            'exit-code-timeout-seconds': 180,
            'kill': 'qdel ${job_id}',
            'check-alive': 'qstat ${job_id}',
            'job-id-regex': '(\\d+)',
        }
    }

    def __init__(
        self,
        local_out_dir,
        max_concurrent_tasks=CromwellBackendBase.DEFAULT_CONCURRENT_JOB_LIMIT,
        soft_glob_output=False,
        local_hash_strat=CromwellBackendLocal.DEFAULT_LOCAL_HASH_STRAT,
        pbs_queue=None,
        pbs_extra_param=None,
    ):
        super().__init__(
            local_out_dir=local_out_dir,
            backend_name=BACKEND_PBS,
            max_concurrent_tasks=max_concurrent_tasks,
            soft_glob_output=soft_glob_output,
            local_hash_strat=local_hash_strat,
        )
        self.merge_backend(CromwellBackendPBS.TEMPLATE_BACKEND)
        self.backend_config.pop('submit-docker')

        if pbs_queue:
            self.default_runtime_attributes['pbs_queue'] = pbs_queue
        if pbs_extra_param:
            self.default_runtime_attributes['pbs_extra_param'] = pbs_extra_param



"""This does not cover all CLI parameters defined in caper/caper_args.py.
gcp (Google Cloud Platform) backend is tested here with server/client functions.
"""
import os
import time

import pytest
from autouri import AutoURI

from caper.cli import main as cli_main
from caper.cromwell_rest_api import CromwellRestAPI
from caper.wdl_parser import WDLParser

from .example_wdl import make_directory_with_wdls

TIMEOUT_SERVER_SPIN_UP = 500
TIMEOUT_SERVER_RUN_WORKFLOW = 960


@pytest.mark.google_cloud
@pytest.mark.integration
def test_server_client(
    tmp_path,
    gcs_root,
    ci_prefix,
    cromwell,
    womtool,
    gcp_prj,
    gcp_service_account_key_json,
    debug_caper,
):
    """Test server, client stuffs
    """
    # server command line
    server_port = 8015

    out_gcs_bucket = os.path.join(gcs_root, 'caper_out', ci_prefix)
    tmp_gcs_bucket = os.path.join(gcs_root, 'caper_tmp')

    cmd = ['server']
    cmd += ['--local-loc-dir', str(tmp_path / 'tmp_dir')]
    cmd += ['--backend', 'gcp']
    if gcp_service_account_key_json:
        cmd += ['--gcp-service-account-key-json', gcp_service_account_key_json]
    cmd += ['--gcp-prj', gcp_prj]
    cmd += ['--gcp-zones', 'us-west1-a,us-west1-b']
    cmd += ['--gcp-out-dir', out_gcs_bucket]
    cmd += ['--gcp-loc-dir', tmp_gcs_bucket]
    cmd += ['--cromwell-stdout', str(tmp_path / 'cromwell_stdout.o')]
    cmd += ['--db', 'in-memory']
    cmd += ['--db-timeout', '500000']
    cmd += ['--file-db', str(tmp_path / 'file_db_prefix')]
    cmd += ['--max-concurrent-tasks', '2']
    cmd += ['--max-concurrent-workflows', '2']
    cmd += ['--disable-call-caching']
    cmd += ['--local-hash-strat', 'path']
    cmd += ['--local-out-dir', str(tmp_path / 'out_dir')]
    cmd += ['--cromwell', cromwell]
    cmd += ['--java-heap-server', '8G']
    cmd += ['--port', str(server_port)]
    if debug_caper:
        cmd += ['--debug']
    print(' '.join(cmd))

    try:
        th = cli_main(cmd, nonblocking_server=True)

        # wait until server is ready to take submissions
        t_start = time.time()
        while th.status is None:
            time.sleep(1)
            if time.time() - t_start > TIMEOUT_SERVER_SPIN_UP:
                raise TimeoutError('Timed out waiting for Cromwell server spin-up.')

        # prepare WDLs and input JSON, imports to be submitted
        make_directory_with_wdls(str(tmp_path))
        wdl = tmp_path / 'main.wdl'
        inputs = tmp_path / 'inputs.json'
        p = WDLParser(str(wdl))
        imports = p.zip_subworkflows(str(tmp_path / 'imports.zip'))

        # test "submit" with on_hold
        cmd = ['submit', str(wdl)]
        if gcp_service_account_key_json:
            cmd += ['--gcp-service-account-key-json', gcp_service_account_key_json]
        cmd += ['--port', str(server_port)]
        cmd += ['--inputs', str(inputs)]
        cmd += ['--imports', str(imports)]
        cmd += ['--gcp-zones', 'us-west1-a,us-west1-b']
        cmd += ['--gcp-loc-dir', tmp_gcs_bucket]
        cmd += ['--ignore-womtool']
        cmd += ['--java-heap-womtool', '2G']
        cmd += ['--max-retries', '1']
        cmd += ['--docker', 'ubuntu:latest']
        cmd += ['--backend', 'gcp']
        cmd += ['--hold']
        if debug_caper:
            cmd += ['--debug']
        cli_main(cmd)

        time.sleep(10)

        # find workflow ID
        cra = CromwellRestAPI(hostname='localhost', port=server_port)
        workflow_id = cra.find(['*'])[0]['id']

        m = cra.get_metadata([workflow_id])[0]
        assert m['status'] == 'On Hold'

        # unhold it
        cmd = ['unhold', workflow_id]
        cmd += ['--port', str(server_port)]
        cli_main(cmd)

        time.sleep(5)

        m = cra.get_metadata([workflow_id])[0]
        assert m['status'] in ('Submitted', 'Running')

        t_start = time.time()
        while True:
            time.sleep(5)
            m = cra.get_metadata([workflow_id])[0]
            workflow_root = m.get('workflowRoot')
            if workflow_root:
                metadata_json_file = os.path.join(workflow_root, 'metadata.json')
            else:
                metadata_json_file = None
            print('polling: ', workflow_id, m['status'], metadata_json_file)

            if m['status'] in ('Failed', 'Succeeded'):
                if AutoURI(metadata_json_file).exists:
                    break
            elif metadata_json_file:
                assert not AutoURI(metadata_json_file).exists

            if time.time() - t_start > TIMEOUT_SERVER_RUN_WORKFLOW:
                raise TimeoutError('Timed out waiting for workflow being done.')

    finally:
        # all done. so stop the server
        if th:
            th.stop()
            th.join()



import argparse
from configparser import DuplicateOptionError
from textwrap import dedent

import pytest

from caper.arg_tool import read_from_conf, update_parsers_defaults_with_conf

CONF_CONTENTS = dedent(
    """\
    param-wo-default="please_remove_double_quote"
    param-w-type-wo-default='4.0'
    param-w-type-wo-default2="5.0"
    param_w_type_wo_default3=
    param-w-int-default=10
    param-w-int-default3=
    flag-w-default=True
    flag-w-default2='False'
    flag-wo-default='FALSE'
    flag-wo-default2="True"
    to_be-replaced=1
"""
)


CONF_CONTENTS_DUPLICATE_ENTRY = dedent(
    """\
    shared-param=200
    shared-param=400
    uniq-param-a=2000
    uniq-param-c=4
"""
)


CONF_CONTENTS_FOR_SUBPARSER = dedent(
    """\
    shared-param=600
    uniq-param-a=2000
    uniq-param-b=4000
    uniq-param-c=4
"""
)


@pytest.fixture
def parser_wo_subparsers():
    parser = argparse.ArgumentParser()
    parser.add_argument('--param-wo-default')
    parser.add_argument('--param-w-type-wo-default', type=float)
    parser.add_argument('--param-w-type-wo-default2', type=float)
    parser.add_argument('--param-w-type-wo-default3', type=float)
    parser.add_argument('--param-w-int-default', default=2)
    parser.add_argument('--param-w-int-default2', default=4)
    parser.add_argument('--param-w-int-default3', default=6)
    parser.add_argument('--flag-w-default', action='store_true', default=False)
    parser.add_argument('--flag-w-default2', action='store_true', default=False)
    parser.add_argument('--flag-wo-default', action='store_true')
    parser.add_argument('--flag-wo-default2', action='store_true')
    return parser


@pytest.fixture
def parser_with_subparsers():
    parser = argparse.ArgumentParser()
    subparser = parser.add_subparsers(dest='action')

    p_sub_a = subparser.add_parser('a')
    p_sub_b = subparser.add_parser('b')

    # two subparsers will have shared and original parameters
    p_sub_a.add_argument('--shared-param', default=2)
    p_sub_a.add_argument('--uniq-param-a', default=20)

    p_sub_b.add_argument('--shared-param', default=4.0)
    p_sub_b.add_argument('--uniq-param-b', default=40.0)
    p_sub_b.add_argument('--uniq-param-c', type=int)

    return parser, [p_sub_a, p_sub_b]


def test_read_from_conf(tmp_path):
    c = tmp_path / 'c1.conf'
    c.write_text(CONF_CONTENTS)

    d1 = read_from_conf(
        c, no_strip_quote=False, conf_key_map={'to_be_replaced': 'replaced_key'}
    )
    assert d1['param_wo_default'] == 'please_remove_double_quote'
    assert d1['param_w_type_wo_default'] == '4.0'
    assert d1['param_w_type_wo_default2'] == '5.0'
    assert 'param_w_type_wo_default3' not in d1
    assert d1['param_w_int_default'] == '10'
    assert 'param_w_int_default3' not in d1
    assert d1['flag_w_default'] == 'True'
    assert d1['flag_w_default2'] == 'False'
    assert d1['flag_wo_default'] == 'FALSE'
    assert d1['flag_wo_default2'] == 'True'
    assert d1['replaced_key'] == '1'
    assert 'to_be-replaced' not in d1

    d2 = read_from_conf(c, no_strip_quote=True)
    assert d2['param_wo_default'] == '"please_remove_double_quote"'
    assert d2['param_w_type_wo_default2'] == '"5.0"'
    assert d2['flag_w_default2'] == '\'False\''
    assert d2['flag_wo_default'] == '\'FALSE\''
    assert d2['flag_wo_default2'] == '"True"'

    c2 = tmp_path / 'c2.conf'
    c2.write_text(CONF_CONTENTS_DUPLICATE_ENTRY)

    with pytest.raises(DuplicateOptionError):
        d2 = read_from_conf(c2)


def test_update_parsers_defaults_with_conf(tmp_path, parser_wo_subparsers):
    """Check if this function correctly updates argparse parser's
    default values.
    """
    val_type = {'param_w_type_wo_default2': float}
    val_default = {'param_w_type_wo_default3': 'hello', 'param_w_int_default3': 50}

    p1 = parser_wo_subparsers
    c1 = tmp_path / 'c1.conf'

    # can mix up _ and -
    c1.write_text(CONF_CONTENTS)
    d1 = update_parsers_defaults_with_conf(
        parsers=[p1], conf_file=str(c1), val_type=val_type, val_default=val_default
    )

    assert p1.get_default('param_wo_default') == 'please_remove_double_quote'
    assert p1.get_default('param_w_type_wo_default') == '4.0'
    assert p1.get_default('param_w_type_wo_default2') == 5.0
    assert p1.get_default('param_w_type_wo_default3') == 'hello'
    assert p1.get_default('param_w_int_default') == 10
    assert p1.get_default('param_w_int_default2') == 4
    assert p1.get_default('param_w_int_default3') == 50
    assert p1.get_default('flag_w_default')
    assert not p1.get_default('flag_w_default2')
    assert not p1.get_default('flag_wo_default')
    assert p1.get_default('flag_wo_default2')

    assert d1['param_wo_default'] == 'please_remove_double_quote'
    assert d1['param_w_type_wo_default'] == '4.0'
    assert d1['param_w_type_wo_default2'] == 5.0
    assert d1['param_w_type_wo_default3'] == 'hello'
    assert d1['param_w_int_default'] == 10
    assert 'param_w_int_default2' not in d1
    assert d1['param_w_int_default3'] == 50
    assert d1['flag_w_default']
    assert not d1['flag_w_default2']
    assert not d1['flag_wo_default']
    assert d1['flag_wo_default2']


def test_update_parsers_defaults_with_conf_with_subparsers(
    tmp_path, parser_with_subparsers
):
    """Check if this function correctly updates argparse parser's
    default values.
    """
    p, subparsers = parser_with_subparsers
    c1 = tmp_path / 'c1.conf'

    # can mix up _ and -
    c1.write_text(CONF_CONTENTS_FOR_SUBPARSER)
    d = update_parsers_defaults_with_conf(parsers=subparsers, conf_file=str(c1))
    args_a, _ = p.parse_known_args(['a'])
    args_b, _ = p.parse_known_args(['b'])

    assert args_a.shared_param == 600.0
    assert args_a.uniq_param_a == 2000

    assert args_b.shared_param == 600.0
    assert args_b.uniq_param_b == 4000.0
    # cannot parse "type" from argparse
    assert args_b.uniq_param_c == 4

    assert d['shared_param'] == 600.0
    assert d['uniq_param_a'] == 2000
    assert d['uniq_param_b'] == 4000.0
    assert d['uniq_param_c'] == '4'



import json
import os

from caper.caper_labels import CaperLabels


def test_create_file(tmp_path):
    cl = CaperLabels()

    backend = 'my_backend'

    custom_labels = tmp_path / 'my_custom_labels.json'
    custom_labels_dict = {'hello': 'world', 'good': {'bye': 'bro'}}
    custom_labels.write_text(json.dumps(custom_labels_dict, indent=4))

    str_label = 'my:str?label*'
    user = 'my_user'
    basename = 'my_basename.json'

    f = cl.create_file(
        directory=str(tmp_path),
        backend=backend,
        custom_labels=str(custom_labels),
        str_label=str_label,
        user=user,
        basename=basename,
    )

    with open(f) as fp:
        d = json.loads(fp.read())

    assert d[CaperLabels.KEY_CAPER_BACKEND] == backend
    assert d['hello'] == 'world'
    assert d['good']['bye'] == 'bro'
    assert d[CaperLabels.KEY_CAPER_STR_LABEL] == 'my_str_label_'
    assert d[CaperLabels.KEY_CAPER_USER] == user
    assert os.path.basename(f) == basename
    assert os.path.dirname(f) == str(tmp_path)



import io
import json
import logging
import os
import re
from collections import defaultdict

import humanfriendly
import numpy as np
import pandas as pd
from autouri import GCSURI, AbsPath, AutoURI, URIBase

from .dict_tool import recurse_dict_value

logger = logging.getLogger(__name__)


def get_workflow_root_from_call(call):
    call_root = call.get('callRoot')
    if call_root:
        return '/'.join(call_root.split('/')[:-1])


def get_workflow_id_from_workflow_root(workflow_root):
    if workflow_root:
        return workflow_root.split('/')[-1]


def parse_cromwell_disks(s):
    """Parses Cromwell's disks in runtime attribute.
    """
    if s:
        m = re.findall(r'(\d+)', s)
        if m:
            return int(m[0]) * 1024 * 1024 * 1024


def parse_cromwell_memory(s):
    """Parses Cromwell's memory runtime attribute.
    """
    if s:
        return humanfriendly.parse_size(s)


def convert_type_np_to_py(o):
    """Convert numpy type to Python type.
    """
    if isinstance(o, np.generic):
        return o.item()
    raise TypeError


class CromwellMetadata:
    DEFAULT_METADATA_BASENAME = 'metadata.json'
    DEFAULT_GCP_MONITOR_STAT_METHODS = ('mean', 'std', 'max', 'min', 'last')

    def __init__(self, metadata):
        """Parses metadata JSON (dict) object or file.
        """
        if isinstance(metadata, dict):
            self._metadata = metadata
        elif isinstance(metadata, CromwellMetadata):
            self._metadata = metadata._metadata
        else:
            s = AutoURI(metadata).read()
            self._metadata = json.loads(s)

    @property
    def data(self):
        return self._metadata

    @property
    def metadata(self):
        return self._metadata

    @property
    def workflow_id(self):
        return self._metadata.get('id')

    @property
    def workflow_status(self):
        return self._metadata.get('status')

    @property
    def workflow_root(self):
        if 'workflowRoot' in self._metadata:
            return self._metadata['workflowRoot']
        else:
            workflow_roots = [
                get_workflow_root_from_call(call) for _, call, _ in self.recursed_calls
            ]
            common_root = os.path.commonprefix(workflow_roots)
            if common_root:
                guessed_workflow_id = get_workflow_id_from_workflow_root(common_root)
                if guessed_workflow_id == self.workflow_id:
                    return common_root
                logger.error(
                    'workflowRoot not found in metadata JSON. '
                    'Tried to guess from callRoot of each call but failed.'
                )

    @property
    def failures(self):
        return self._metadata.get('failures')

    @property
    def calls(self):
        return self._metadata.get('calls')

    @property
    def recursed_calls(self):
        """Returns a generator for tuples.

        Tuple:
            call_name:
                Call's name. i.e. key in the original metadata JSON's `calls` dict.
            call:
                Call object. i.e. value in the original metadata JSON's `calls` dict.
            parent_call_names:
                Tuple of Parent call's names.
        """
        return self.recurse_calls(
            lambda call_name, call, parent_call_names: (
                call_name,
                call,
                parent_call_names,
            )
        )

    def recurse_calls(self, fn_call, parent_call_names=tuple()):
        """Recurse on tasks in metadata.

        Args:
            fn_call:
                Function to be called recursively for each call (task).
                This function should take the following three arguments.
                    call_name:
                        Call's name. i.e. key in the original metadata JSON's `calls` dict.
                    call:
                        Call object. i.e. value in the original metadata JSON's `calls` dict.
                    parent_call_names:
                        Tuple of Parent call's names.
                        e.g. (..., great grand parent, grand parent, parent, ...)
        Returns:
            Generator object for all calls.
        """
        if not self.calls:
            return

        for call_name, call_list in self.calls.items():
            for call in call_list:
                if 'subWorkflowMetadata' in call:
                    subworkflow = call['subWorkflowMetadata']
                    subworkflow_metadata = CromwellMetadata(subworkflow)
                    yield from subworkflow_metadata.recurse_calls(
                        fn_call, parent_call_names=parent_call_names + (call_name,)
                    )
                else:
                    yield fn_call(call_name, call, parent_call_names)

    def write_on_workflow_root(self, basename=DEFAULT_METADATA_BASENAME):
        """Update metadata JSON file on metadata's output root directory.
        """
        root = self.workflow_root

        if root:
            metadata_file = os.path.join(root, basename)

            AutoURI(metadata_file).write(json.dumps(self._metadata, indent=4) + '\n')
            logger.info('Wrote metadata file. {f}'.format(f=metadata_file))

            return metadata_file

    def troubleshoot(self, show_completed_task=False, show_stdout=False):
        """Troubleshoots a workflow.
        Also, finds failure reasons and prints out STDERR and STDOUT.

        Args:
            show_completed_task:
                Show STDERR/STDOUT of completed tasks.
            show_stdout:
                Show failed task's STDOUT along with STDERR.
        Return:
            result:
                Troubleshooting report as a plain string.
        """
        result = '* Started troubleshooting workflow: id={id}, status={status}\n'.format(
            id=self.workflow_id, status=self.workflow_status
        )

        if self.workflow_status == 'Succeeded':
            result += '* Workflow ran Successfully.\n'

        else:
            if self.failures:
                result += '* Found failures JSON object.\n{s}\n'.format(
                    s=json.dumps(self.failures, indent=4)
                )

            def troubleshoot_call(call_name, call, parent_call_names):
                """Returns troubleshooting help message.
                """
                nonlocal show_completed_task
                nonlocal show_stdout
                status = call.get('executionStatus')
                shard_index = call.get('shardIndex')
                rc = call.get('returnCode')
                job_id = call.get('jobId')
                stdout = call.get('stdout')
                stderr = call.get('stderr')
                run_start = None
                run_end = None
                for event in call.get('executionEvents', []):
                    if event['description'].startswith('Running'):
                        run_start = event['startTime']
                        run_end = event['endTime']
                        break

                help_msg = ''
                if show_completed_task or status not in ('Done', 'Succeeded'):
                    help_msg += (
                        '\n==== NAME={name}, STATUS={status}, PARENT={p}\n'
                        'SHARD_IDX={shard_idx}, RC={rc}, JOB_ID={job_id}\n'
                        'START={start}, END={end}\n'
                        'STDOUT={stdout}\nSTDERR={stderr}\n'.format(
                            name=call_name,
                            status=status,
                            p=','.join(parent_call_names),
                            start=run_start,
                            end=run_end,
                            shard_idx=shard_index,
                            rc=rc,
                            job_id=job_id,
                            stdout=stdout,
                            stderr=stderr,
                        )
                    )
                    if stderr:
                        if AutoURI(stderr).exists:
                            help_msg += 'STDERR_CONTENTS=\n{s}\n'.format(
                                s=AutoURI(stderr).read()
                            )
                    if show_stdout and stdout:
                        if AutoURI(stdout).exists:
                            help_msg += 'STDOUT_CONTENTS=\n{s}\n'.format(
                                s=AutoURI(stdout).read()
                            )

                return help_msg

            result += '* Recursively finding failures in calls (tasks)...\n'
            for help_msg in self.recurse_calls(troubleshoot_call):
                result += help_msg

        return result

    def gcp_monitor(
        self,
        task_name=None,
        excluded_cols=(0,),
        stat_methods=DEFAULT_GCP_MONITOR_STAT_METHODS,
    ):
        """Recursively parse task(call)'s `monitoringLog`
        (`monitoring.log` in task's execution directory)
        generated by `monitoring_script` defined in workflow options.
        This feature is gcp backend only.
        Check the following for details.
        https://cromwell.readthedocs.io/en/stable/wf_options/Google/#google-pipelines-api-workflow-options

        This functions calculates mean/max/min/last of each column in `monitoring.log` and return
        them with task's input file sizes.

        Args:
            task_name:
                If defined, limit analysis to this task only.
            excluded_cols:
                List of 0-based indices of excluded columns. There will be no mean/max/min
                calculation for these excluded columns.
                col-0 (1st column) is excluded by default since it's usually a column
                for timestamps.
            stat_methods:
                List/tuple of stat method strings.
                Except for `last`, any method of pandas.DataFrame can be used for stat_methods.
                e.g. `mean`, `max`, `min`, ...
                `last` is to get the last element in data, which usually means the latest data.
                Some methods in pandas.DataFrame will return `nan` if the number of data row is
                too small (e.g. `std` requires more than one data row).
        Returns:
            List of mean/std/max/min/last of columns along with size of input files.
            Note that
                - None will be returned if there are no data in the file.
                    - This means that the log file exists but there is no data in it.
                - `shard_idx` is -1 for non-scattered tasks.
                - Dot notation (.) will be used for task_name of subworkflow's task.

            Result format:
            [
                {
                    'workflow_id': WORKFLOW_ID,
                    'task_name': TASK_NAME,
                    'status': TASK_STATUS,
                    'shard_idx': SHARD_INDEX,
                    'attempt': ATTEMPT_RETRIAL,
                    'mean': {
                        COL1_NAME: MEAN_OF_COL1,
                        COL2_NAME: MEAN_OF_COL2,
                        ...
                    },
                    'gcp_instance': {
                        'cpu': NUM_CPU,
                        'disk': DISK_SIZE_USED,
                        'mem': TOTAL_MEMORY_IN_BYTES,
                    },
                    'stats': {
                        'std': {
                            COL1_NAME: STD_OF_COL1,
                            COL2_NAME: STD_OF_COL2,
                            ...
                        },
                        'max': {
                            COL1_NAME: MAX_OF_COL1,
                            COL2_NAME: MAX_OF_COL2,
                            ...
                        },
                        'min': {
                            COL1_NAME: MIN_OF_COL1,
                            COL2_NAME: MIN_OF_COL2,
                            ...
                        },
                        'last': {
                            COL1_NAME: LAST_ENTRY_OF_COL1,
                            COL2_NAME: LAST_ENTRY_OF_COL2,
                            ...
                        },
                    },
                    'input_file_sizes': {
                        INPUT1: [
                            SIZE_OF_FILE1_IN_INPUT1,
                            SIZE_OF_FILE2_IN_INPUT1,
                            ...
                        ],
                        INPUT2: [
                            SIZE_OF_FILE1_IN_INPUT2,
                            ...
                        ],
                        ...
                    },
                },
                ...
            ]
        """
        file_size_cache = {}
        workflow_id = self.workflow_id

        def gcp_monitor_call(call_name, call, parent_call_names):
            nonlocal excluded_cols
            nonlocal stat_methods
            nonlocal file_size_cache
            nonlocal workflow_id
            nonlocal task_name

            if task_name and task_name != call_name:
                return

            monitoring_log = call.get('monitoringLog')
            if monitoring_log is None:
                return
            if not GCSURI(monitoring_log).is_valid:
                # This feature is for GCSURI only.
                return
            if not GCSURI(monitoring_log).exists:
                # Workaround for Cromwell-52's bug.
                # Call-cached task has `monitoringLog`, but it does not exist.
                return

            dataframe = pd.read_csv(
                io.StringIO(GCSURI(monitoring_log).read()), delimiter='\t'
            )
            rt_attrs = call.get('runtimeAttributes')

            data = {
                'workflow_id': workflow_id,
                'task_name': call_name,
                'shard_idx': call.get('shardIndex'),
                'status': call.get('executionStatus'),
                'attempt': call.get('attempt'),
                'instance': {
                    'cpu': int(rt_attrs.get('cpu')),
                    'disk': parse_cromwell_disks(rt_attrs.get('disks')),
                    'mem': parse_cromwell_memory(rt_attrs.get('memory')),
                },
                'stats': {s: {} for s in stat_methods},
                'input_file_sizes': defaultdict(list),
            }
            for i, col_name in enumerate(dataframe.columns):
                if i in excluded_cols:
                    continue
                for stat_method in stat_methods:
                    if dataframe.empty:
                        val = None
                    elif stat_method == 'last':
                        last_idx = dataframe.tail(1).index.item()
                        val = dataframe[col_name][last_idx]
                    else:
                        val = getattr(dataframe[col_name], stat_method)()
                    data['stats'][stat_method][col_name] = val

            for input_name, input_value in sorted(call['inputs'].items()):
                file_sizes_dict = data['input_file_sizes']

                def add_to_input_files_if_valid(file):
                    nonlocal file_size_cache
                    nonlocal file_sizes_dict
                    nonlocal input_name

                    if GCSURI(file).is_valid:
                        file_size = file_size_cache.get(file)
                        if file_size is None:
                            file_size = GCSURI(file).size
                            file_size_cache[file] = file_size
                        file_sizes_dict[input_name].append(file_size)

                recurse_dict_value(input_value, add_to_input_files_if_valid)

            return data

        result = list(self.recurse_calls(gcp_monitor_call))

        # a bit hacky way to recursively convert numpy type into python type
        json_str = json.dumps(result, default=convert_type_np_to_py)
        return json.loads(json_str)

    def cleanup(
        self, dry_run=False, num_threads=URIBase.DEFAULT_NUM_THREADS, no_lock=False
    ):
        """Cleans up workflow's root output directory.

        Args:
            dry_run:
                Dry-run mode.
            num_threads:
                For outputs on cloud buckets only.
                Number of threads for deleting individual outputs on cloud buckets in parallel.
                Generates one client per thread. This works like `gsutil -m rm -rf`.
            no_lock:
                No file locking.
        """
        root = self.workflow_root
        if not root:
            logger.error(
                'workflow\'s root directory cannot be found in metadata JSON. '
                'Cannot proceed to cleanup outputs.'
            )
            return

        if AbsPath(root).is_valid:
            # num_threads is not available for AbsPath().rmdir()
            AbsPath(root).rmdir(dry_run=dry_run, no_lock=no_lock)
        else:
            AutoURI(root).rmdir(
                dry_run=dry_run, no_lock=no_lock, num_threads=num_threads
            )



#!/usr/bin/env python3
import copy
import csv
import json
import logging
import os
import re
import sys

from autouri import GCSURI, AutoURI

from . import __version__ as version
from .caper_args import ResourceAnalysisReductionMethod, get_parser_and_defaults
from .caper_client import CaperClient, CaperClientSubmit
from .caper_init import init_caper_conf
from .caper_labels import CaperLabels
from .caper_runner import CaperRunner
from .cromwell_backend import (
    BACKEND_ALIAS_LOCAL,
    BACKEND_LOCAL,
    CromwellBackendDatabase,
)
from .cromwell_metadata import CromwellMetadata
from .dict_tool import flatten_dict
from .resource_analysis import LinearResourceAnalysis
from .server_heartbeat import ServerHeartbeat

logger = logging.getLogger(__name__)


DEFAULT_DB_FILE_PREFIX = 'caper_file_db'
DEFAULT_SERVER_HEARTBEAT_FILE = '~/.caper/default_server_heartbeat'
USER_INTERRUPT_WARNING = (
    '\n\n'
    '*** DO NOT CTRL+C MULTIPLE TIMES! ***\n'
    '*** OR CAPER WILL NOT BE ABLE TO STOP CROMWELL AND RUNNING WORKFLOWS/TASKS ***'
    '\n\n'
)
REGEX_DELIMITER_PARAMS = r',| '
PRINT_ROW_DELIMITER = '\t'


def get_abspath(path):
    """Get abspath from a string.
    This function is mainly used to make a command line argument an abspath
    since AutoURI module only works with abspath and full URIs
    (e.g. /home/there, gs://here/there).
    For example, "caper run toy.wdl --docker ubuntu:latest".
    AutoURI cannot recognize toy.wdl on CWD as a file path.
    It should be converted to an abspath first.
    To do so, use this function for local file path strings only (e.g. toy.wdl).
    Do not use this function for other non-local-path strings (e.g. --docker).
    """
    if path:
        if not AutoURI(path).is_valid:
            return os.path.abspath(os.path.expanduser(path))
    return path


def print_version(parser, args):
    if args.version:
        print(version)
        parser.exit()


def init_logging(args):
    if args.debug:
        log_level = 'DEBUG'
    else:
        log_level = 'INFO'
    logging.basicConfig(
        level=log_level, format='%(asctime)s|%(name)s|%(levelname)s| %(message)s'
    )
    # suppress filelock logging
    logging.getLogger('filelock').setLevel('CRITICAL')


def init_autouri(args):
    if hasattr(args, 'use_gsutil_for_s3'):
        GCSURI.init_gcsuri(use_gsutil_for_s3=args.use_gsutil_for_s3)


def check_flags(args):
    singularity_flag = False
    docker_flag = False

    if hasattr(args, 'singularity') and args.singularity is not None:
        singularity_flag = True
        if args.singularity.endswith(('.wdl', '.cwl')):
            raise ValueError(
                '--singularity ate up positional arguments (e.g. WDL, CWL). '
                'Define --singularity at the end of command line arguments. '
                'singularity={p}'.format(p=args.singularity)
            )

    if hasattr(args, 'docker') and args.docker is not None:
        docker_flag = True
        if args.docker.endswith(('.wdl', '.cwl')):
            raise ValueError(
                '--docker ate up positional arguments (e.g. WDL, CWL). '
                'Define --docker at the end of command line arguments. '
                'docker={p}'.format(p=args.docker)
            )
        if hasattr(args, 'soft_glob_output') and args.soft_glob_output:
            raise ValueError(
                '--soft-glob-output and --docker are mutually exclusive. '
                'Delocalization from docker container will fail '
                'for soft-linked globbed outputs.'
            )

    if singularity_flag and docker_flag:
        raise ValueError('--docker and --singularity are mutually exclusive.')


def check_dirs(args):
    """Convert local directories (local_out_dir, local_loc_dir) to absolute ones.
    Also, if temporary/cache directory is not defined for each storage,
    then append ".caper_tmp" on output directory and use it.
    """
    if hasattr(args, 'local_out_dir'):
        args.local_out_dir = get_abspath(args.local_out_dir)
        if not args.local_loc_dir:
            args.local_loc_dir = os.path.join(
                args.local_out_dir, CaperRunner.DEFAULT_LOC_DIR_NAME
            )
    else:
        if not args.local_loc_dir:
            args.local_loc_dir = os.path.join(
                os.getcwd(), CaperRunner.DEFAULT_LOC_DIR_NAME
            )

    args.local_loc_dir = get_abspath(args.local_loc_dir)

    if hasattr(args, 'gcp_out_dir'):
        if args.gcp_out_dir and not args.gcp_loc_dir:
            args.gcp_loc_dir = os.path.join(
                args.gcp_out_dir, CaperRunner.DEFAULT_LOC_DIR_NAME
            )

    if hasattr(args, 'aws_out_dir'):
        if args.aws_out_dir and not args.aws_loc_dir:
            args.aws_loc_dir = os.path.join(
                args.aws_out_dir, CaperRunner.DEFAULT_LOC_DIR_NAME
            )


def check_db_path(args):
    if hasattr(args, 'db') and args.db == CromwellBackendDatabase.DB_FILE:
        args.file_db = get_abspath(args.file_db)

        if not args.file_db:
            prefix = DEFAULT_DB_FILE_PREFIX
            if hasattr(args, 'inputs') and args.inputs:
                prefix += '_' + os.path.splitext(os.path.basename(args.inputs))[0]

            args.file_db = os.path.join(args.local_out_dir, prefix)


def check_backend(args):
    """Check if local backend is in lower cases.
    "Local" should be capitalized. i.e. local -> Local.
    BACKEND_LOCAL is Local.
    BACKEND_ALIAS_LOCAL is local.
    """
    if hasattr(args, 'backend') and args.backend == BACKEND_ALIAS_LOCAL:
        args.backend = BACKEND_LOCAL


def runner(args, nonblocking_server=False):
    if args.gcp_zones:
        args.gcp_zones = re.split(REGEX_DELIMITER_PARAMS, args.gcp_zones)
    if args.memory_retry_error_keys:
        args.memory_retry_error_keys = re.split(
            REGEX_DELIMITER_PARAMS, args.memory_retry_error_keys
        )

    c = CaperRunner(
        local_loc_dir=args.local_loc_dir,
        local_out_dir=args.local_out_dir,
        default_backend=args.backend,
        gcp_loc_dir=args.gcp_loc_dir,
        aws_loc_dir=args.aws_loc_dir,
        gcp_service_account_key_json=get_abspath(args.gcp_service_account_key_json),
        cromwell=get_abspath(args.cromwell),
        womtool=get_abspath(getattr(args, 'womtool', None)),
        disable_call_caching=args.disable_call_caching,
        max_concurrent_workflows=args.max_concurrent_workflows,
        memory_retry_error_keys=args.memory_retry_error_keys,
        max_concurrent_tasks=args.max_concurrent_tasks,
        soft_glob_output=args.soft_glob_output,
        local_hash_strat=args.local_hash_strat,
        db=args.db,
        db_timeout=args.db_timeout,
        file_db=args.file_db,
        mysql_db_ip=args.mysql_db_ip,
        mysql_db_port=args.mysql_db_port,
        mysql_db_user=args.mysql_db_user,
        mysql_db_password=args.mysql_db_password,
        mysql_db_name=args.mysql_db_name,
        postgresql_db_ip=args.postgresql_db_ip,
        postgresql_db_port=args.postgresql_db_port,
        postgresql_db_user=args.postgresql_db_user,
        postgresql_db_password=args.postgresql_db_password,
        postgresql_db_name=args.postgresql_db_name,
        gcp_prj=args.gcp_prj,
        use_google_cloud_life_sciences=args.use_google_cloud_life_sciences,
        gcp_region=args.gcp_region,
        gcp_zones=args.gcp_zones,
        gcp_call_caching_dup_strat=args.gcp_call_caching_dup_strat,
        gcp_out_dir=args.gcp_out_dir,
        aws_batch_arn=args.aws_batch_arn,
        aws_region=args.aws_region,
        aws_out_dir=args.aws_out_dir,
        aws_call_caching_dup_strat=args.aws_call_caching_dup_strat,
        slurm_partition=getattr(args, 'slurm_partition', None),
        slurm_account=getattr(args, 'slurm_account', None),
        slurm_extra_param=getattr(args, 'slurm_extra_param', None),
        sge_pe=getattr(args, 'sge_pe', None),
        sge_queue=getattr(args, 'sge_queue', None),
        sge_extra_param=getattr(args, 'sge_extra_param', None),
        pbs_queue=getattr(args, 'pbs_queue', None),
        pbs_extra_param=getattr(args, 'pbs_extra_param', None),
    )

    if args.action == 'run':
        subcmd_run(c, args)

    elif args.action == 'server':
        return subcmd_server(c, args, nonblocking=nonblocking_server)

    else:
        raise ValueError('Unsupported runner action {act}'.format(act=args.action))


def client(args):
    sh = None
    if not args.no_server_heartbeat:
        sh = ServerHeartbeat(
            heartbeat_file=args.server_heartbeat_file,
            heartbeat_timeout=args.server_heartbeat_timeout,
        )
    if args.action == 'submit':
        if args.gcp_zones:
            args.gcp_zones = re.split(REGEX_DELIMITER_PARAMS, args.gcp_zones)

        c = CaperClientSubmit(
            local_loc_dir=args.local_loc_dir,
            gcp_loc_dir=args.gcp_loc_dir,
            aws_loc_dir=args.aws_loc_dir,
            gcp_service_account_key_json=get_abspath(args.gcp_service_account_key_json),
            server_hostname=args.hostname,
            server_port=args.port,
            server_heartbeat=sh,
            womtool=get_abspath(args.womtool),
            use_google_cloud_life_sciences=args.use_google_cloud_life_sciences,
            gcp_zones=args.gcp_zones,
            slurm_partition=args.slurm_partition,
            slurm_account=args.slurm_account,
            slurm_extra_param=args.slurm_extra_param,
            sge_pe=args.sge_pe,
            sge_queue=args.sge_queue,
            sge_extra_param=args.sge_extra_param,
            pbs_queue=args.pbs_queue,
            pbs_extra_param=args.pbs_extra_param,
        )
        subcmd_submit(c, args)

    else:
        c = CaperClient(
            local_loc_dir=args.local_loc_dir,
            gcp_loc_dir=args.gcp_loc_dir,
            aws_loc_dir=args.aws_loc_dir,
            gcp_service_account_key_json=get_abspath(args.gcp_service_account_key_json),
            server_hostname=args.hostname,
            server_port=args.port,
            server_heartbeat=sh,
        )
        if args.action == 'abort':
            subcmd_abort(c, args)
        elif args.action == 'unhold':
            subcmd_unhold(c, args)
        elif args.action == 'list':
            subcmd_list(c, args)
        elif args.action == 'metadata':
            subcmd_metadata(c, args)
        elif args.action in ('troubleshoot', 'debug'):
            subcmd_troubleshoot(c, args)
        elif args.action == 'gcp_monitor':
            subcmd_gcp_monitor(c, args)
        elif args.action == 'gcp_res_analysis':
            subcmd_gcp_res_analysis(c, args)
        elif args.action == 'cleanup':
            subcmd_cleanup(c, args)
        else:
            raise ValueError('Unsupported client action {act}'.format(act=args.action))


def subcmd_server(caper_runner, args, nonblocking=False):
    """
    Args:
        nonblocking:
            Make this function return a Thread object
            instead of blocking (Thread.join()).
            Also writes Cromwell's STDOUT to sys.stdout
            instead of a file (args.cromwell_stdout).
    """
    sh = None
    if not args.no_server_heartbeat:
        sh = ServerHeartbeat(
            heartbeat_file=args.server_heartbeat_file,
            heartbeat_timeout=args.server_heartbeat_timeout,
        )

    args_from_cli = {
        'default_backend': args.backend,
        'server_port': args.port,
        'server_heartbeat': sh,
        'custom_backend_conf': get_abspath(args.backend_file),
        'embed_subworkflow': True,
        'auto_write_metadata': not args.disable_auto_write_metadata,
        'java_heap_server': args.java_heap_server,
        'dry_run': args.dry_run,
    }

    if nonblocking:
        return caper_runner.server(fileobj_stdout=sys.stdout, **args_from_cli)

    cromwell_stdout = get_abspath(args.cromwell_stdout)
    with open(cromwell_stdout, 'w') as f:
        try:
            thread = caper_runner.server(fileobj_stdout=f, **args_from_cli)
            if thread:
                thread.join()
                thread.stop(wait=True)
                if thread.returncode:
                    logger.error('Check stdout in {file}'.format(file=cromwell_stdout))

        except KeyboardInterrupt:
            logger.error(USER_INTERRUPT_WARNING, exc_info=True)


def subcmd_run(caper_runner, args):
    cromwell_stdout = get_abspath(args.cromwell_stdout)

    with open(cromwell_stdout, 'w') as f:
        try:
            thread = caper_runner.run(
                backend=args.backend,
                wdl=get_abspath(args.wdl),
                inputs=get_abspath(args.inputs),
                options=get_abspath(args.options),
                labels=get_abspath(args.labels),
                imports=get_abspath(args.imports),
                metadata_output=get_abspath(args.metadata_output),
                str_label=args.str_label,
                docker=args.docker,
                singularity=args.singularity,
                singularity_cachedir=args.singularity_cachedir,
                no_build_singularity=args.no_build_singularity,
                custom_backend_conf=get_abspath(args.backend_file),
                max_retries=args.max_retries,
                memory_retry_multiplier=args.memory_retry_multiplier,
                gcp_monitoring_script=args.gcp_monitoring_script,
                ignore_womtool=args.ignore_womtool,
                no_deepcopy=args.no_deepcopy,
                fileobj_stdout=f,
                fileobj_troubleshoot=sys.stdout,
                java_heap_run=args.java_heap_run,
                java_heap_womtool=args.java_heap_womtool,
                dry_run=args.dry_run,
            )
            if thread:
                thread.join()
                thread.stop(wait=True)
                if thread.returncode:
                    logger.error('Check stdout in {file}'.format(file=cromwell_stdout))

        except KeyboardInterrupt:
            logger.error(USER_INTERRUPT_WARNING, exc_info=True)


def subcmd_submit(caper_client, args):
    caper_client.submit(
        wdl=get_abspath(args.wdl),
        backend=args.backend,
        inputs=get_abspath(args.inputs),
        options=get_abspath(args.options),
        labels=get_abspath(args.labels),
        imports=get_abspath(args.imports),
        str_label=args.str_label,
        docker=args.docker,
        singularity=args.singularity,
        singularity_cachedir=args.singularity_cachedir,
        no_build_singularity=args.no_build_singularity,
        max_retries=args.max_retries,
        memory_retry_multiplier=args.memory_retry_multiplier,
        gcp_monitoring_script=args.gcp_monitoring_script,
        ignore_womtool=args.ignore_womtool,
        no_deepcopy=args.no_deepcopy,
        hold=args.hold,
        java_heap_womtool=args.java_heap_womtool,
        dry_run=args.dry_run,
    )


def subcmd_abort(caper_client, args):
    caper_client.abort(args.wf_id_or_label)


def subcmd_unhold(caper_client, args):
    caper_client.unhold(args.wf_id_or_label)


def subcmd_list(caper_client, args):
    workflows = caper_client.list(
        args.wf_id_or_label, exclude_subworkflow=not args.show_subworkflow
    )

    try:
        writer = csv.writer(sys.stdout, delimiter=PRINT_ROW_DELIMITER)

        formats = args.format.split(',')
        writer.writerow(formats)

        if workflows is None:
            return
        for w in workflows:
            row = []
            workflow_id = w.get('id')
            parent_workflow_id = w.get('parentWorkflowId')

            if args.hide_result_before is not None:
                if (
                    w.get('submission')
                    and w.get('submission') <= args.hide_result_before
                ):
                    continue
            for f in formats:
                if f == 'workflow_id':
                    row.append(str(workflow_id))
                elif f == 'str_label':
                    if 'labels' in w and CaperLabels.KEY_CAPER_STR_LABEL in w['labels']:
                        lbl = w['labels'][CaperLabels.KEY_CAPER_STR_LABEL]
                    else:
                        lbl = None
                    row.append(str(lbl))
                elif f == 'user':
                    if 'labels' in w and CaperLabels.KEY_CAPER_USER in w['labels']:
                        lbl = w['labels'][CaperLabels.KEY_CAPER_USER]
                    else:
                        lbl = None
                    row.append(str(lbl))
                elif f == 'parent':
                    row.append(str(parent_workflow_id))
                else:
                    row.append(str(w.get(f)))
            writer.writerow(row)

    except BrokenPipeError:
        logger.debug('Ignored BrokenPipeError.')


def subcmd_metadata(caper_client, args):
    m = caper_client.metadata(
        wf_ids_or_labels=args.wf_id_or_label, embed_subworkflow=True
    )
    if not m:
        raise ValueError('Found no workflow matching with search query.')
    elif len(m) > 1:
        raise ValueError('Found multiple workflow matching with search query.')

    print(json.dumps(m[0], indent=4))


def get_single_cromwell_metadata_obj(caper_client, args, subcmd):
    if not args.wf_id_or_label:
        raise ValueError(
            'Define at least one metadata JSON file or '
            'a search query for workflow ID/string label '
            'if there is a running Caper server.'
        )
    elif len(args.wf_id_or_label) > 1:
        raise ValueError(
            'Multiple files/queries are not allowed for {subcmd}. '
            'Define one metadata JSON file or a search query '
            'for workflow ID/string label.'.format(subcmd=subcmd)
        )

    metadata_file = AutoURI(get_abspath(args.wf_id_or_label[0]))

    if metadata_file.exists:
        metadata = json.loads(metadata_file.read())
    else:
        metadata_objs = caper_client.metadata(
            wf_ids_or_labels=args.wf_id_or_label, embed_subworkflow=True
        )
        if len(metadata_objs) > 1:
            raise ValueError('Found multiple workflows matching with search query.')
        elif len(metadata_objs) == 0:
            raise ValueError('Found no workflow matching with search query.')
        metadata = metadata_objs[0]

    return CromwellMetadata(metadata)


def split_list_into_file_and_non_file(lst):
    """Returns tuple of (list of existing files, list of non-file strings)
    """
    files = []
    non_files = []

    for maybe_file in lst:
        if AutoURI(get_abspath(maybe_file)).exists:
            files.append(maybe_file)
        else:
            non_files.append(maybe_file)

    return files, non_files


def get_multi_cromwell_metadata_objs(caper_client, args):
    if not args.wf_id_or_label:
        raise ValueError(
            'Define at least one metadata JSON file or '
            'a search query for workflow ID/string label '
            'if there is a running Caper server.'
        )

    files, non_files = split_list_into_file_and_non_file(args.wf_id_or_label)

    all_metadata = []
    for file in files:
        metadata = json.loads(AutoURI(get_abspath(file)).read())
        all_metadata.append(metadata)

    if non_files:
        all_metadata.extend(
            caper_client.metadata(wf_ids_or_labels=non_files, embed_subworkflow=True)
        )

    if not all_metadata:
        raise ValueError('Found no metadata/workflow matching with search query.')
    return [CromwellMetadata(m) for m in all_metadata]


def subcmd_troubleshoot(caper_client, args):
    cm = get_single_cromwell_metadata_obj(caper_client, args, 'troubleshoot/debug')
    sys.stdout.write(
        cm.troubleshoot(
            show_completed_task=args.show_completed_task, show_stdout=args.show_stdout
        )
    )


def subcmd_gcp_monitor(caper_client, args):
    """Prints out monitoring result either in a TSV format or in a JSON one.

    TSV format:
        TSV will be a flattened JSON with dot notation.
        The last column in the header is `input_file_name_size_pairs`.
        As its name implies, contents for this last columns are dynamic in length.

    JSON format:
        See description at CromwellMetadata.gcp_monitor.__doc__.
        Use a JSON format instead to get more detailed information.
    """
    all_metadata = get_multi_cromwell_metadata_objs(caper_client, args)
    writer = csv.writer(sys.stdout, delimiter=PRINT_ROW_DELIMITER)

    result = []
    for metadata in all_metadata:
        result.extend(metadata.gcp_monitor())

    if args.json_format:
        print(json.dumps(result, indent=4))
    else:
        # input_file_sizes is dynamic in length so exclude and then put it back
        first_data = copy.deepcopy(result[0])
        first_data.pop('input_file_sizes')
        header = list(flatten_dict(first_data, reducer='.').keys())
        header += ['input_file_var_size_pairs']
        writer.writerow(header)

        for task_data in result:
            input_file_sizes = task_data.pop('input_file_sizes')
            row = [str(v) for v in flatten_dict(task_data).values()]

            # append `input_file_sizes` data which couldn't be cleanly
            # flattened by flatten_dict()
            for key, file_sizes in input_file_sizes.items():
                for i, file_size in enumerate(file_sizes):
                    if len(file_sizes) == 1:
                        row.append(key)
                    else:
                        row.append(key + '[{idx}]'.format(idx=i))
                    row.append(str(file_size))

            writer.writerow(row)


def read_json(json_file):
    if json_file:
        json_contents = AutoURI(get_abspath(json_file)).read()
        return json.loads(json_contents)


def subcmd_gcp_res_analysis(caper_client, args):
    """Solves linear regression problem to find coeffs and intercept
    to help optimizing resources for a task based on task's input file size.

    Prints out found coeffs and intercept along with raw dataset (x, y).
        - x: input file sizes for a task
        - y: resources (max_mem, max_disk) taken for a task
    """
    all_metadata = get_multi_cromwell_metadata_objs(caper_client, args)

    res_analysis = LinearResourceAnalysis()
    res_analysis.collect_resource_data(all_metadata)

    result = res_analysis.analyze(
        in_file_vars=read_json(args.in_file_vars_def_json),
        reduce_in_file_vars=getattr(
            ResourceAnalysisReductionMethod, args.reduce_in_file_vars
        ).value,
        target_resources=args.target_resources,
        plot_pdf=get_abspath(args.plot_pdf),
    )
    print(json.dumps(result, indent=4))


def subcmd_cleanup(caper_client, args):
    """Cleanup outputs of a workflow.
    """
    cm = get_single_cromwell_metadata_obj(caper_client, args, 'cleanup')
    cm.cleanup(dry_run=not args.delete, num_threads=args.num_threads, no_lock=True)
    if not args.delete:
        logger.warning(
            'Use --delete to DELETE ALL OUTPUTS of this workflow. '
            'This action is NOT REVERSIBLE. Use this at your own risk.'
        )


def main(args=None, nonblocking_server=False):
    """
    Args:
        args:
            List of command line arguments.
            If defined use it instead of sys.argv.
        nonblocking_server:
            "server" subcommand will return a Thread object
            instead of waiting (Thread.join()).
    """
    parser, _ = get_parser_and_defaults()

    if args is None and len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    known_args, _ = parser.parse_known_args(args)
    check_flags(known_args)
    print_version(parser, known_args)

    parsed_args = parser.parse_args(args)
    init_logging(parsed_args)
    init_autouri(parsed_args)

    check_dirs(parsed_args)
    check_db_path(parsed_args)
    check_backend(parsed_args)

    if parsed_args.action == 'init':
        init_caper_conf(parsed_args.conf, parsed_args.platform)

    elif parsed_args.action in ('run', 'server'):
        return runner(parsed_args, nonblocking_server=nonblocking_server)
    else:
        client(parsed_args)


if __name__ == '__main__':
    main()



from textwrap import dedent

from caper.dict_tool import (
    dict_to_dot_str,
    flatten_dict,
    merge_dict,
    split_dict,
    unflatten_dict,
)


def test_merge_dict():
    d1 = {
        'flagstat_qc': {'rep1': {'read1': 100}, 'rep2': {'read2': 400}},
        'etc': {'samstat_qc': {'rep1': {'unmapped': 500, 'mapped': 600}}},
    }
    d2 = {
        'flagstat_qc': {'rep1': {'read2': 200}, 'rep2': {'read1': 300}},
        'etc': {'samstat_qc': {'rep2': {'unmapped': 700, 'mapped': 800}}},
        'idr_qc': {'qc_test1': 900},
    }

    assert merge_dict(d1, d2) == {
        'flagstat_qc': {
            'rep1': {'read1': 100, 'read2': 200},
            'rep2': {'read1': 300, 'read2': 400},
        },
        'etc': {
            'samstat_qc': {
                'rep1': {'unmapped': 500, 'mapped': 600},
                'rep2': {'unmapped': 700, 'mapped': 800},
            }
        },
        'idr_qc': {'qc_test1': 900},
    }


def test_flatten_dict():
    d = {
        'flagstat_qc': {
            'rep1': {'read1': 100, 'read2': 200},
            'rep2': {'read1': 300, 'read2': 400},
        },
        'rep': 1,
    }
    assert flatten_dict(d) == {
        ('flagstat_qc', 'rep1', 'read1'): 100,
        ('flagstat_qc', 'rep1', 'read2'): 200,
        ('flagstat_qc', 'rep2', 'read1'): 300,
        ('flagstat_qc', 'rep2', 'read2'): 400,
        ('rep',): 1,
    }


def test_unflatten_dict():
    d_f = {
        ('flagstat_qc', 'rep1', 'read1'): 100,
        ('flagstat_qc', 'rep1', 'read2'): 200,
        ('flagstat_qc', 'rep2', 'read1'): 300,
        ('flagstat_qc', 'rep2', 'read2'): 400,
        ('rep',): 1,
    }
    assert unflatten_dict(d_f) == {
        'flagstat_qc': {
            'rep1': {'read1': 100, 'read2': 200},
            'rep2': {'read1': 300, 'read2': 400},
        },
        'rep': 1,
    }


def test_split_dict():
    d = {
        'flagstat_qc': {
            'rep1': {'read1': 100, 'read2': 200},
            'rep2': {'read1': 300, 'read2': 400},
        },
        'etc': {
            'samstat_qc': {
                'rep1': {'unmapped': 500, 'mapped': 600},
                'rep2': {'unmapped': 700, 'mapped': 800},
            }
        },
        'idr_qc': {'qc_test1': 900},
    }
    splits = split_dict(d, ('replicate', r'^rep\d+$'))
    splits_ref = [
        {'idr_qc': {'qc_test1': 900}},
        {
            'flagstat_qc': {'read1': 100, 'read2': 200},
            'etc': {'samstat_qc': {'unmapped': 500, 'mapped': 600}},
            'replicate': 'rep1',
        },
        {
            'flagstat_qc': {'read1': 300, 'read2': 400},
            'etc': {'samstat_qc': {'unmapped': 700, 'mapped': 800}},
            'replicate': 'rep2',
        },
    ]
    assert splits == splits_ref


def test_dict_to_dot_str():
    d = {
        'rankDir': 'TD',
        'start': '[shape=Mdiamond]',
        'end': '[shape=Msquare]',
        'subgraph cluster_rep1': {
            'style': 'filled',
            'color': 'mistyrose',
            'label': '"Replicate 1"',
        },
        'subgraph cluster_rep2': {
            'style': 'filled',
            'color': 'azure',
            'label': '"Replicate 2"',
        },
        'a0 -> b0': None,
        'c0 -> d0': None,
    }
    dot = dict_to_dot_str(d, parent_key='digraph D', indent=' ' * 4)
    ref = dedent(
        """\
        digraph D {
            rankDir = TD;
            start = [shape=Mdiamond];
            end = [shape=Msquare];
            subgraph cluster_rep1 {
                style = filled;
                color = mistyrose;
                label = "Replicate 1";
            }
            subgraph cluster_rep2 {
                style = filled;
                color = azure;
                label = "Replicate 2";
            }
            a0 -> b0;
            c0 -> d0;
        }
    """
    )
    assert dot == ref



import logging
import os
from datetime import datetime

from autouri import GCSURI, S3URI, AbsPath, AutoURI

from .cromwell_backend import BACKEND_AWS, BACKEND_GCP

logger = logging.getLogger(__name__)


class CaperBase:
    ENV_GOOGLE_APPLICATION_CREDENTIALS = 'GOOGLE_APPLICATION_CREDENTIALS'
    DEFAULT_LOC_DIR_NAME = '.caper_tmp'

    def __init__(
        self,
        local_loc_dir=None,
        gcp_loc_dir=None,
        aws_loc_dir=None,
        gcp_service_account_key_json=None,
    ):
        """Manages work/cache/temp directories for localization on the following
        storages:
            - Local*: Local path -> local_loc_dir**
            - gcp: GCS bucket path -> gcp_loc_dir
            - aws: S3 bucket path -> aws_loc_dir

        * Note that it starts with capital L, which is a default backend of Cromwell's
        default configuration file (application.conf).
        ** /tmp is not recommended. This directory is very important to store
        intermediate files used by Cromwell/AutoURI (file transfer/localization).

        Also manages Google Cloud auth (key JSON file) since both Caper client/server
        require permission to access to storage.

        Args:
            local_loc_dir:
                Local cache directory to store files localized for local backends.
                Unlike other two directories. This directory is also used to make a
                working directory to store intermediate files to run Cromwell.
                e.g. backend.conf and workflow_opts.json.
            gcp_loc_dir:
                GCS cache directory to store files localized on GCS for gcp backend.
            aws_loc_dir:
                S3 cache directory to store files localized on S3 for aws backend.
            gcp_service_account_key_json:
                Google Cloud service account for authentication.
                This service account should have enough permission to storage.
        """
        if local_loc_dir is None:
            local_loc_dir = os.path.join(os.getcwd(), CaperBase.DEFAULT_LOC_DIR_NAME)

        if not AbsPath(local_loc_dir).is_valid:
            raise ValueError(
                'local_loc_dir should be a valid local abspath. {f}'.format(
                    f=local_loc_dir
                )
            )
        if gcp_loc_dir and not GCSURI(gcp_loc_dir).is_valid:
            raise ValueError(
                'gcp_loc_dir should be a valid GCS path. {f}'.format(f=gcp_loc_dir)
            )
        if aws_loc_dir and not S3URI(aws_loc_dir).is_valid:
            raise ValueError(
                'aws_loc_dir should be a valid S3 path. {f}'.format(f=aws_loc_dir)
            )

        self._local_loc_dir = local_loc_dir
        self._gcp_loc_dir = gcp_loc_dir
        self._aws_loc_dir = aws_loc_dir

        self._set_env_gcp_app_credentials(gcp_service_account_key_json)

    def _set_env_gcp_app_credentials(
        self,
        gcp_service_account_key_json=None,
        env_name=ENV_GOOGLE_APPLICATION_CREDENTIALS,
    ):
        """Initalizes environment for authentication (VM instance/storage).

        Args:
            gcp_service_account_key_json:
                Secret key JSON file for auth.
                This service account should have full permission to storage and
                VM instance.
                Environment variable GOOGLE_APPLICATION_CREDENTIALS will be
                updated with this.
        """
        if gcp_service_account_key_json:
            gcp_service_account_key_json = os.path.expanduser(
                gcp_service_account_key_json
            )
            if env_name in os.environ:
                auth_file = os.environ[env_name]
                if not os.path.samefile(auth_file, gcp_service_account_key_json):
                    logger.warning(
                        'Env var {env} does not match with '
                        'gcp_service_account_key_json. '
                        'Using application default credentials? '.format(env=env_name)
                    )
            logger.debug(
                'Adding GCP service account key JSON {key} to '
                'env var {env}'.format(key=gcp_service_account_key_json, env=env_name)
            )
            os.environ[env_name] = gcp_service_account_key_json

    def localize_on_backend(self, f, backend, recursive=False, make_md5_file=False):
        """Localize a file according to the chosen backend.
        Each backend has its corresponding storage.
            - gcp -> GCS bucket path (starting with gs://)
            - aws -> S3 bucket path (starting with s3://)
            - All others (based on local backend) -> local storage

        If contents of input JSON changes due to recursive localization (deepcopy)
        then a new temporary file suffixed with backend type will be written on loc_prefix.
        For example, /somewhere/test.json -> gs://example-tmp-gcs-bucket/somewhere/test.gcs.json

        loc_prefix will be one of the cache directories according to the backend type
            - gcp -> gcp_loc_dir
            - aws -> aws_loc_dir
            - all others (local) -> local_loc_dir

        Args:
            f:
                File to be localized.
            backend:
                Backend to localize file f on.
            recursive:
                Recursive localization (deepcopy).
                All files (if value is valid path/URI string) in JSON/CSV/TSV
                will be localized together with file f.
            make_md5_file:
                Make .md5 file for localized files. This is for local only since
                GCS/S3 bucket paths already include md5 hash information in their metadata.

        Returns:
            localized URI.
        """
        if backend == BACKEND_GCP:
            loc_prefix = self._gcp_loc_dir
        elif backend == BACKEND_AWS:
            loc_prefix = self._aws_loc_dir
        else:
            loc_prefix = self._local_loc_dir

        return AutoURI(f).localize_on(
            loc_prefix, recursive=recursive, make_md5_file=make_md5_file
        )

    def localize_on_backend_if_modified(
        self, f, backend, recursive=False, make_md5_file=False
    ):
        """Wrapper for localize_on_backend.

        If localized file is not modified due to recursive localization,
        then it means that localization for such file was redundant.
        So returns the original file instead of a redundantly localized one.
        We can check if file is modifed or not by looking at their basename.
        Modified localized file has a suffix of the target storage. e.g. .s3.
        """
        f_loc = self.localize_on_backend(
            f=f, backend=backend, recursive=recursive, make_md5_file=make_md5_file
        )

        if AutoURI(f).basename == AutoURI(f_loc).basename:
            return f
        return f_loc

    def create_timestamped_work_dir(self, prefix=''):
        """Creates/returns a local temporary directory on self._local_work_dir.

        Args:
            prefix:
                Prefix for timstamped directory.
                Directory name will be self._tmpdir / prefix / timestamp.
        """
        timestamp = datetime.now().strftime('%Y%m%d_%H%M%S_%f')
        work_dir = os.path.join(self._local_loc_dir, prefix, timestamp)
        os.makedirs(work_dir, exist_ok=True)
        logger.info(
            'Creating a timestamped temporary directory. {d}'.format(d=work_dir)
        )

        return work_dir

    def get_loc_dir(self, backend):
        """Get localization directory for a backend.
        """
        if backend == BACKEND_GCP:
            return self._gcp_loc_dir
        elif backend == BACKEND_AWS:
            return self._aws_loc_dir
        else:
            return self._local_loc_dir



"""WDLs used for testing.

To test many functions recursively working with subworkflows
these WDLs have the following structure:

    main.wdl (imports sub.wdl)
        sub/
            sub.wdl (imports sub_sub.wdl)
            sub/
                sub_sub.wdl (imports nothing)
        inputs.json (inputs JSON file)
"""
import json
import os
from textwrap import dedent

from autouri import AutoURI

WRONG_WDL = dedent(
    """\
    version 1.0

    workflwwwwwwwwwowwwww main {
    }
"""
)

MAIN_WDL = dedent(
    """\
    version 1.0
    import "sub/sub.wdl" as sub

    workflow main {
        meta {
            key1: "val1"
            key2: "val2"
        }
        parameter_meta {
            input_s: {
                key1: "val1"
            }
            input_i: {
                key1: "val1"
            }
        }
        input {
            String input_s
            Int input_i = 1
        }

        call t1
        call sub.sub

        output {
            File out1 = t1.out
            File out_sub = sub.out
        }
    }

    task t1 {
        command {
            echo 1 > out.txt
            sleep 10
        }
        output {
            File out = 'out.txt'
        }
    }
"""
)

MAIN_INPUTS = {'main.input_s': 'a'}

MAIN_WDL_META_DICT = {'key1': 'val1', 'key2': 'val2'}


MAIN_WDL_PARAMETER_META_DICT = {
    'input_s': {'key1': 'val1'},
    'input_i': {'key1': 'val1'},
}

SUB_WDL = dedent(
    """\
    version 1.0
    import "sub/sub_sub.wdl" as sub_sub

    workflow sub {
        call t2
        call sub_sub.sub_sub

        output {
            File out = t2.out
            File out_sub = sub_sub.out
        }
    }

    task t2 {
        command {
            echo 2 > out2.txt
        }
        output {
            File out = 'out2.txt'
        }
    }
"""
)

SUB_WDL_TO_FAIL = dedent(
    """\
    version 1.0
    import "sub/sub_sub.wdl" as sub_sub

    workflow sub {
        call t2_failing
        call sub_sub.sub_sub

        output {
            File out = t2_failing.out
            File out_sub = sub_sub.out
        }
    }

    task t2_failing {
        command {
            echo 2 > out2.txt
            INTENTED_ERROR
        }
        output {
            File out = 'out2.txt'
        }
    }
"""
)

SUB_SUB_WDL = dedent(
    """\
    version 1.0

    workflow sub_sub {
        call t3
        output {
            File out = t3.out
        }
    }

    task t3 {
        command {
            echo 3 > out3.txt
        }
        output {
            File out = 'out3.txt'
        }
    }
"""
)


def make_directory_with_wdls(directory, no_sub_wdl=False):
    """
    Run Cromwell with WDLs:
    main + 1 sub + 1 sub's sub.

    Returns:
        Created root directory
    """
    main_inputs = os.path.join(directory, 'inputs.json')
    AutoURI(main_inputs).write(json.dumps(MAIN_INPUTS, indent=4))

    main_wdl = os.path.join(directory, 'main.wdl')
    AutoURI(main_wdl).write(MAIN_WDL)

    if not no_sub_wdl:
        sub_wdl = os.path.join(directory, 'sub', 'sub.wdl')
        AutoURI(sub_wdl).write(SUB_WDL)

        sub_sub_wdl = os.path.join(directory, 'sub', 'sub', 'sub_sub.wdl')
        AutoURI(sub_sub_wdl).write(SUB_SUB_WDL)


def make_directory_with_failing_wdls(directory, no_sub_wdl=False):
    """
    Run Cromwell with WDLs:
    main + 1 sub (supposed to fail) + 1 sub's sub.

    Returns:
        Created root directory
    """
    main_inputs = os.path.join(directory, 'inputs.json')
    AutoURI(main_inputs).write(json.dumps(MAIN_INPUTS, indent=4))

    main_wdl = os.path.join(directory, 'main.wdl')
    AutoURI(main_wdl).write(MAIN_WDL)

    if not no_sub_wdl:
        sub_wdl = os.path.join(directory, 'sub', 'sub.wdl')
        AutoURI(sub_wdl).write(SUB_WDL_TO_FAIL)

        sub_sub_wdl = os.path.join(directory, 'sub', 'sub', 'sub_sub.wdl')
        AutoURI(sub_sub_wdl).write(SUB_SUB_WDL)



import json
import os
import sys
import time

import pytest

from caper.cromwell import Cromwell, WomtoolValidationFailed
from caper.cromwell_rest_api import CromwellRestAPI
from caper.wdl_parser import WDLParser

from .example_wdl import WRONG_WDL, make_directory_with_wdls

BACKEND_CONF_CONTENTS = """
backend {{
  providers {{
    Local {{
      config {{
        root = {root}
      }}
    }}
  }}
}}
"""

TIMEOUT_SERVER_SPIN_UP = 200
TIMEOUT_SERVER_RUN_WORKFLOW = 960


def test_validate(tmp_path, cromwell, womtool):
    c = Cromwell(cromwell=cromwell, womtool=womtool)

    wdl = tmp_path / 'wrong.wdl'
    wdl.write_text(WRONG_WDL)
    with pytest.raises(WomtoolValidationFailed):
        c.validate(str(wdl))

    make_directory_with_wdls(str(tmp_path / 'successful'))
    wdl = tmp_path / 'successful' / 'main.wdl'
    inputs = tmp_path / 'successful' / 'inputs.json'
    c.validate(str(wdl), str(inputs))

    # zip subworkflows for later use
    p = WDLParser(str(wdl))
    imports = p.zip_subworkflows(str(tmp_path / 'imports.zip'))

    # test with imports.zip
    make_directory_with_wdls(str(tmp_path / 'wo_sub_wdls'), no_sub_wdl=True)
    wdl = tmp_path / 'wo_sub_wdls' / 'main.wdl'
    inputs = tmp_path / 'wo_sub_wdls' / 'inputs.json'
    c.validate(str(wdl), str(inputs), imports)


def test_run(tmp_path, cromwell, womtool):
    fileobj_stdout = sys.stdout

    c = Cromwell(cromwell=cromwell, womtool=womtool)

    make_directory_with_wdls(str(tmp_path))

    o_dir = tmp_path / 'output'
    o_dir.mkdir()
    work_dir = tmp_path / 'work_dir'
    work_dir.mkdir()

    backend_conf = tmp_path / 'backend.conf'
    backend_conf.write_text(BACKEND_CONF_CONTENTS.format(root=o_dir))

    try:
        th = c.run(
            backend_conf=str(backend_conf),
            wdl=str(tmp_path / 'main.wdl'),
            inputs=str(tmp_path / 'inputs.json'),
            metadata=str(tmp_path / 'metadata.json'),
            fileobj_stdout=fileobj_stdout,
            work_dir=work_dir,
            cwd=str(tmp_path),
        )
    finally:
        th.join()
    assert th.returncode == 0

    # check if metadata.json is written on both specified location
    # (tmp_path/metadata.json) and workflow's root directory
    metadata_dict = th.returnvalue
    root_dir = metadata_dict['workflowRoot']

    with open(os.path.join(root_dir, 'metadata.json')) as fp:
        metadata_contents_on_root = fp.read()
    metadata_dict_on_root = json.loads(metadata_contents_on_root)

    assert metadata_dict == metadata_dict_on_root
    # check if backend_conf's change of root directory worked
    assert root_dir.startswith(str(o_dir))

    # zip subworkflows for later use
    p = WDLParser(str(tmp_path / 'main.wdl'))
    imports = p.zip_subworkflows(str(tmp_path / 'imports.zip'))

    # test without sub WDLs but with imports.zip
    # test run without work_dir
    make_directory_with_wdls(str(tmp_path / 'wo_sub_wdls'), no_sub_wdl=True)

    try:
        th = c.run(
            wdl=str(tmp_path / 'wo_sub_wdls' / 'main.wdl'),
            inputs=str(tmp_path / 'wo_sub_wdls' / 'inputs.json'),
            imports=imports,
            fileobj_stdout=fileobj_stdout,
            cwd=str(tmp_path / 'wo_sub_wdls'),
        )
    finally:
        th.join()
    assert th.returncode == 0


def test_server(tmp_path, cromwell, womtool):
    """Test Cromwell.server() method, which returns a Thread object.
    """
    server_port = 8005
    fileobj_stdout = sys.stdout

    c = Cromwell(cromwell=cromwell, womtool=womtool)

    o_dir = tmp_path / 'output'
    o_dir.mkdir()

    backend_conf = tmp_path / 'backend.conf'
    backend_conf.write_text(BACKEND_CONF_CONTENTS.format(root=o_dir))

    is_server_started = False

    def on_server_start():
        nonlocal is_server_started
        is_server_started = True

    workflow_id = None
    is_workflow_done = False

    def on_status_change(metadata):
        nonlocal workflow_id
        nonlocal is_workflow_done

        if metadata:
            if metadata['id'] == workflow_id:
                if metadata['status'] in ('Succeeded', 'Failed'):
                    is_workflow_done = True

    # also tests two callback functions
    try:
        th = c.server(
            server_port=server_port,
            backend_conf=str(backend_conf),
            embed_subworkflow=True,
            fileobj_stdout=fileobj_stdout,
            on_server_start=on_server_start,
            on_status_change=on_status_change,
            cwd=str(tmp_path),
        )
        assert th.status is None

        # wait until server is ready to take submissions
        t_start = time.time()
        while not is_server_started:
            time.sleep(1)
            if time.time() - t_start > TIMEOUT_SERVER_SPIN_UP:
                raise TimeoutError('Timed out waiting for Cromwell server spin-up.')

        # another way of checking server is started
        assert th.status

        # make WDLs and imports
        wdl = tmp_path / 'main.wdl'
        make_directory_with_wdls(str(tmp_path))
        # zip subworkflows for later use
        p = WDLParser(str(wdl))
        imports = p.zip_subworkflows(str(tmp_path / 'imports.zip'))

        cra = CromwellRestAPI(hostname='localhost', port=server_port)
        r = cra.submit(
            source=str(wdl), dependencies=imports, inputs=str(tmp_path / 'inputs.json')
        )
        workflow_id = r['id']

        t_start = time.time()
        while not is_workflow_done:
            time.sleep(1)
            print('polling: ', workflow_id, is_workflow_done)
            if time.time() - t_start > TIMEOUT_SERVER_RUN_WORKFLOW:
                raise TimeoutError('Timed out waiting for workflow being done.')

        metadata = cra.get_metadata([workflow_id], embed_subworkflow=True)[0]

        # check if metadata JSON is written on workflow's root directory.
        root_dir = metadata['workflowRoot']
        metadata_file = os.path.join(root_dir, 'metadata.json')
        assert os.path.exists(metadata_file)

        # check if subworkflow is embedded.
        with open(metadata_file) as fp:
            metadata_from_file = json.loads(fp.read())
        assert metadata == metadata_from_file

    finally:
        th.stop()
        th.join()



import fnmatch
import io
import logging
from uuid import UUID

import requests
from requests.exceptions import ConnectionError, HTTPError

from .cromwell_metadata import CromwellMetadata

logger = logging.getLogger(__name__)


def requests_error_handler(func):
    """Re-raise ConnectionError with help message.
    Continue on HTTP 404 error (server is on but workflow doesn't exist).
    Otherwise, re-raise from None to hide nested tracebacks.
    """

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)

        except HTTPError as err:
            if err.response.status_code == 404:
                logger.error("Workflow doesn't seem to exist.")
                return

            message = (
                '{err}\n\n'
                'Cromwell server is on but got an HTTP error other than 404. '
            ).format(err=err)
            raise HTTPError(message) from None

        except ConnectionError as err:
            message = (
                '{err}\n\n'
                'Failed to connect to Cromwell server. '
                'Check if Caper server is running. '
                'Also check if hostname and port are correct. '
                'method={method}, '
                'url={url}'.format(
                    err=err, method=err.request.method, url=err.request.url
                )
            )
            raise ConnectionError(message) from None

    return wrapper


def is_valid_uuid(workflow_id, version=4):
    """To validate Cromwell's UUID (lowercase only).
    This does not allow uppercase UUIDs.
    """
    if not isinstance(workflow_id, str):
        return False
    if not workflow_id.islower():
        return False

    try:
        UUID(workflow_id, version=version)
    except ValueError:
        return False
    return True


def has_wildcard(workflow_id_or_label):
    """Check if string or any element in list/tuple has
    a wildcard (? or *).

    Args:
        workflow_id_or_label:
            Workflow ID (str) or label (str).
            Or array (list, tuple) of them.
    """
    if workflow_id_or_label is None:
        return False
    if isinstance(workflow_id_or_label, (list, tuple)):
        for val in workflow_id_or_label:
            if has_wildcard(val):
                return True
        return False
    else:
        return '?' in workflow_id_or_label or '*' in workflow_id_or_label


class CromwellRestAPI:
    QUERY_URL = 'http://{hostname}:{port}'
    ENDPOINT_BACKEND = '/api/workflows/v1/backends'
    ENDPOINT_WORKFLOWS = '/api/workflows/v1/query'
    ENDPOINT_METADATA = '/api/workflows/v1/{wf_id}/metadata'
    ENDPOINT_LABELS = '/api/workflows/v1/{wf_id}/labels'
    ENDPOINT_SUBMIT = '/api/workflows/v1'
    ENDPOINT_ABORT = '/api/workflows/v1/{wf_id}/abort'
    ENDPOINT_RELEASE_HOLD = '/api/workflows/v1/{wf_id}/releaseHold'
    DEFAULT_HOSTNAME = 'localhost'
    DEFAULT_PORT = 8000

    def __init__(
        self, hostname=DEFAULT_HOSTNAME, port=DEFAULT_PORT, user=None, password=None
    ):
        self._hostname = hostname
        self._port = port

        self._user = user
        self._password = password
        self.__init_auth()

    def submit(
        self,
        source,
        dependencies=None,
        inputs=None,
        options=None,
        labels=None,
        on_hold=False,
    ):
        """Submit a workflow.

        Returns:
            JSON Response from POST request submit a workflow
        """
        manifest = {}
        with open(source) as fp:
            manifest['workflowSource'] = io.StringIO(fp.read())
        if dependencies:
            with open(dependencies, 'rb') as fp:
                manifest['workflowDependencies'] = io.BytesIO(fp.read())
        if inputs:
            with open(inputs) as fp:
                manifest['workflowInputs'] = io.StringIO(fp.read())
        else:
            manifest['workflowInputs'] = io.StringIO('{}')
        if options:
            with open(options) as fp:
                manifest['workflowOptions'] = io.StringIO(fp.read())
        if labels:
            with open(labels) as fp:
                manifest['labels'] = io.StringIO(fp.read())
        if on_hold:
            manifest['workflowOnHold'] = True

        r = self.__request_post(CromwellRestAPI.ENDPOINT_SUBMIT, manifest)
        logger.debug('submit: {r}'.format(r=r))
        return r

    def abort(self, workflow_ids=None, labels=None):
        """Abort workflows matching workflow IDs or labels

        Returns:
            List of JSON responses from POST request
            for aborting workflows
        """
        valid_workflow_ids = self.find_valid_workflow_ids(
            workflow_ids=workflow_ids, labels=labels
        )
        if valid_workflow_ids is None:
            return

        result = []
        for workflow_id in valid_workflow_ids:
            r = self.__request_post(
                CromwellRestAPI.ENDPOINT_ABORT.format(wf_id=workflow_id)
            )
            result.append(r)
        logger.debug('abort: {r}'.format(r=result))
        return result

    def release_hold(self, workflow_ids=None, labels=None):
        """Release hold of workflows matching workflow IDs or labels

        Returns:
            List of JSON responses from POST request
            for releasing hold of workflows
        """
        valid_workflow_ids = self.find_valid_workflow_ids(
            workflow_ids=workflow_ids, labels=labels
        )
        if valid_workflow_ids is None:
            return

        result = []
        for workflow_id in valid_workflow_ids:
            r = self.__request_post(
                CromwellRestAPI.ENDPOINT_RELEASE_HOLD.format(wf_id=workflow_id)
            )
            result.append(r)
        logger.debug('release_hold: {r}'.format(r=result))
        return result

    def get_default_backend(self):
        """Retrieve default backend name

        Returns:
            Default backend name
        """
        return self.get_backends()['defaultBackend']

    def get_backends(self):
        """Retrieve available backend names and default backend name

        Returns:
            JSON response with keys "defaultBackend" and "supportedBackends"
            Example: {"defaultBackend":"Local","supportedBackends":
                      ["Local","aws","gcp","pbs","sge","slurm"]}
        """
        return self.__request_get(CromwellRestAPI.ENDPOINT_BACKEND)

    def find_valid_workflow_ids(
        self, workflow_ids=None, labels=None, exclude_subworkflow=True
    ):
        """Checks if workflow ID in `workflow_ids` are already valid UUIDs (without wildcards).
        If so then we don't have to send the server a query to get matching workflow IDs.
        """
        if not labels and workflow_ids and all(is_valid_uuid(i) for i in workflow_ids):
            return workflow_ids
        else:
            workflows = self.find(
                workflow_ids=workflow_ids,
                labels=labels,
                exclude_subworkflow=exclude_subworkflow,
            )
            if not workflows:
                return
            return [w['id'] for w in workflows]

    def get_metadata(self, workflow_ids=None, labels=None, embed_subworkflow=False):
        """Retrieve metadata for workflows matching workflow IDs or labels

        Args:
            workflow_ids:
                List of workflows IDs to find workflows matched.
            labels:
                List of Caper's string labels to find workflows matched.
            embed_subworkflow:
                Recursively embed subworkflow's metadata in main
                workflow's metadata.
                This flag is to mimic behavior of Cromwell run mode with -m.
                Metadata JSON generated with Cromwell run mode
                includes all subworkflows embedded in main workflow's JSON file.
        """
        valid_workflow_ids = self.find_valid_workflow_ids(
            workflow_ids=workflow_ids, labels=labels
        )
        if valid_workflow_ids is None:
            return

        result = []
        for workflow_id in valid_workflow_ids:
            params = {}
            if embed_subworkflow:
                params['expandSubWorkflows'] = True

            m = self.__request_get(
                CromwellRestAPI.ENDPOINT_METADATA.format(wf_id=workflow_id),
                params=params,
            )
            if m:
                cm = CromwellMetadata(m)
                result.append(cm.metadata)
        return result

    def get_labels(self, workflow_id):
        """Get labels JSON for a specified workflow

        Returns:
            Labels JSON for a workflow
        """
        if workflow_id is None or not is_valid_uuid(workflow_id):
            return

        r = self.__request_get(
            CromwellRestAPI.ENDPOINT_LABELS.format(wf_id=workflow_id)
        )
        if r is None:
            return
        return r['labels']

    def get_label(self, workflow_id, key):
        """Get a label for a key in a specified workflow

        Returns:
            Value for a specified key in labels JSON for a workflow
        """
        labels = self.get_labels(workflow_id)
        if labels is None:
            return
        if key in labels:
            return labels[key]

    def update_labels(self, workflow_id, labels):
        """Update labels for a specified workflow with
        a list of (key, val) tuples
        """
        if workflow_id is None or labels is None:
            return
        r = self.__request_patch(
            CromwellRestAPI.ENDPOINT_LABELS.format(wf_id=workflow_id), labels
        )
        logger.debug('update_labels: {r}'.format(r=r))
        return r

    def find_with_wildcard(
        self, workflow_ids=None, labels=None, exclude_subworkflow=True
    ):
        """Retrieves all workflows from Cromwell server.
        And then find matching workflows by ID or labels.
        Wildcards (? and *) are allowed for both parameters.
        """
        result = []

        if not workflow_ids and not labels:
            return result

        resp = self.__request_get(
            CromwellRestAPI.ENDPOINT_WORKFLOWS,
            params={
                'additionalQueryResultFields': 'labels',
                'includeSubworkflows': not exclude_subworkflow,
            },
        )

        if resp and resp['results']:
            for workflow in resp['results']:
                matched = False
                if 'id' not in workflow:
                    continue
                if workflow_ids:
                    for wf_id in workflow_ids:
                        if fnmatch.fnmatchcase(workflow['id'], wf_id):
                            result.append(workflow)
                            matched = True
                            break
                if matched:
                    continue
                if labels and 'labels' in workflow:
                    for k, v in labels:
                        v_ = workflow['labels'].get(k)
                        if not v_:
                            continue
                        if isinstance(v_, str) and isinstance(v, str):
                            # matching with wildcards for str values only
                            if fnmatch.fnmatchcase(v_, v):
                                result.append(workflow)
                                break
                        elif v_ == v:
                            result.append(workflow)
                            break
            logger.debug(
                'find_with_wildcard: workflow_ids={workflow_ids}, '
                'labels={labels}, result={result}'.format(
                    workflow_ids=workflow_ids, labels=labels, result=result
                )
            )

        return result

    def find_by_workflow_ids(self, workflow_ids=None, exclude_subworkflow=True):
        """Finds workflows by exactly matching workflow IDs (UUIDs).
        Does OR search for a list of workflow IDs.
        Invalid UUID in `workflows_ids` will be ignored without warning.
        Wildcards (? and *) are not allowed.

        Args:
            workflow_ids:
                List of workflow ID (UUID) strings.
                Lower-case only (Cromwell uses lower-case UUIDs).
        Returns:
            List of matched workflow JSONs.
        """
        if has_wildcard(workflow_ids):
            raise ValueError(
                'Wildcards are not allowed in workflow_ids. '
                'ids={ids}'.format(ids=workflow_ids)
            )

        result = []
        if workflow_ids:
            # exclude invalid workflow UUIDs.
            workflow_ids = [wf_id for wf_id in workflow_ids if is_valid_uuid(wf_id)]
            resp = self.__request_get(
                CromwellRestAPI.ENDPOINT_WORKFLOWS,
                params={
                    'additionalQueryResultFields': 'labels',
                    'includeSubworkflows': not exclude_subworkflow,
                    'id': workflow_ids,
                },
            )
            if resp and resp['results']:
                result.extend(resp['results'])

            logger.debug(
                'find_by_workflow_ids: workflow_ids={workflow_ids}, '
                'result={result}'.format(workflow_ids=workflow_ids, result=result)
            )

        return result

    def find_by_labels(self, labels=None, exclude_subworkflow=True):
        """Finds workflows by exactly matching labels (key, value) tuples.
        Does OR search for a list of label key/value pairs.
        Wildcards (? and *) are not allowed.

        Args:
            labels:
                List of labels (key/value pairs).
        Returns:
            List of matched workflow JSONs.
        """
        if has_wildcard(labels):
            raise ValueError(
                'Wildcards are not allowed in labels. '
                'labels={labels}'.format(labels=labels)
            )

        result = []
        if labels:
            # reformat labels with `:` notation. exclude pairs with empty value.
            labels = [
                '{key}:{val}'.format(key=key, val=val) for key, val in labels if val
            ]
            resp = self.__request_get(
                CromwellRestAPI.ENDPOINT_WORKFLOWS,
                params={
                    'additionalQueryResultFields': 'labels',
                    'includeSubworkflows': not exclude_subworkflow,
                    'labelor': labels,
                },
            )
            if resp and resp['results']:
                result.extend(resp['results'])

            logger.debug(
                'find_by_labels: labels={labels}, result={result}'.format(
                    labels=labels, result=result
                )
            )

        return result

    def find(self, workflow_ids=None, labels=None, exclude_subworkflow=True):
        """Wrapper for the following three find functions.
        - find_with_wildcard
        - find_by_workflow_ids
        - find_by_labels

        Find workflows by matching workflow IDs or label (key, value) tuples.
        Does OR search for both parameters.
        Wildcards (? and *) in both parameters are allowed but Caper will
        retrieve a list of all workflows, which can lead to HTTP 503 of
        Cromwell server if there are many subworkflows and not `exclude_subworkflow`.

        Args:
            workflow_ids:
                List of workflow ID (UUID) strings.
                Lower-case only.
            labels:
                List of labels (key/value pairs).
            exclude_subworkflow:
                Exclude subworkflows.
        Returns:
            List of matched workflow JSONs.
        """
        wildcard_found_in_workflow_ids = has_wildcard(workflow_ids)
        wildcard_found_in_labels = has_wildcard(
            [val for key, val in labels] if labels else None
        )
        if wildcard_found_in_workflow_ids or wildcard_found_in_labels:
            return self.find_with_wildcard(
                workflow_ids=workflow_ids,
                labels=labels,
                exclude_subworkflow=exclude_subworkflow,
            )

        result = []

        result_by_labels = self.find_by_labels(
            labels=labels, exclude_subworkflow=exclude_subworkflow
        )
        result.extend(result_by_labels)

        workflow_ids_found_by_labels = [workflow['id'] for workflow in result_by_labels]
        result.extend(
            [
                workflow
                for workflow in self.find_by_workflow_ids(
                    workflow_ids=workflow_ids, exclude_subworkflow=exclude_subworkflow
                )
                if workflow['id'] not in workflow_ids_found_by_labels
            ]
        )

        return result

    def __init_auth(self):
        """Init auth object
        """
        if self._user is not None and self._password is not None:
            self._auth = (self._user, self._password)
        else:
            self._auth = None

    @requests_error_handler
    def __request_get(self, endpoint, params=None):
        """GET request

        Returns:
            JSON response
        """
        url = (
            CromwellRestAPI.QUERY_URL.format(hostname=self._hostname, port=self._port)
            + endpoint
        )
        resp = requests.get(
            url, auth=self._auth, params=params, headers={'accept': 'application/json'}
        )
        resp.raise_for_status()
        return resp.json()

    @requests_error_handler
    def __request_post(self, endpoint, manifest=None):
        """POST request

        Returns:
            JSON response
        """
        url = (
            CromwellRestAPI.QUERY_URL.format(hostname=self._hostname, port=self._port)
            + endpoint
        )
        resp = requests.post(
            url, files=manifest, auth=self._auth, headers={'accept': 'application/json'}
        )
        resp.raise_for_status()
        return resp.json()

    @requests_error_handler
    def __request_patch(self, endpoint, data):
        """POST request

        Returns:
            JSON response
        """
        url = (
            CromwellRestAPI.QUERY_URL.format(hostname=self._hostname, port=self._port)
            + endpoint
        )
        resp = requests.patch(
            url,
            data=data,
            auth=self._auth,
            headers={'accept': 'application/json', 'content-type': 'application/json'},
        )
        resp.raise_for_status()
        return resp.json()



import logging
import time
from signal import SIGTERM
from subprocess import PIPE, Popen
from threading import Thread

logger = logging.getLogger(__name__)


def is_fileobj_open(fileobj):
    return fileobj and not getattr(fileobj, 'closed', False)


class NBSubprocThread(Thread):
    DEFAULT_POLL_INTERVAL_SEC = 0.01
    DEFAULT_SUBPROCESS_NAME = 'Subprocess'
    DEFAULT_STOP_SIGNAL = SIGTERM

    def __init__(
        self,
        args,
        cwd=None,
        stdin=None,
        on_poll=None,
        on_stdout=None,
        on_stderr=None,
        on_finish=None,
        poll_interval=DEFAULT_POLL_INTERVAL_SEC,
        quiet=False,
        subprocess_name=DEFAULT_SUBPROCESS_NAME,
    ):
        """Non-blocking STDOUT/STDERR streaming for subprocess.Popen().

        This class makes two daemonized threads for nonblocking
        streaming of STDOUT/STDERR.

        Note that return value of callback functions are updated
        for the following properties:
            - status:
                Updated with return value of on_poll, on_stdout, on_stderr.
                If return value is None then no update.
            - returnvalue:
                Updated with return value of on_finish.
                If return value is None then no update.

        This is useful to check status of the thread and
        get the final return value of the function that this class
        actually runs.

        Args:
            args:
                List of command line arguments.
            cwd:
                subprocess.Popen's cwd.
            stdin:
                subprocess.Popen's stdin.
                Note that subprocess.Popen's stdout/stderr is fixed
                at subprocess.PIPE/subprocess.STDOUT.
            on_poll:
                Callback on every polling.
                If return value is not None then it is used for updating property `status`.
            on_stdout:
                Callback on every non-empty STDOUT line.
                If return value is not None then it is used for updating property `status`.
                This callback function should take one argument:
                    - stdout (str):
                        New incoming STDOUT line string with trailing newline (backslash n).
            on_stderr:
                Callback on every non-empty STDERR line.
                If return value is not None then it is used for updating property `status`.
                This callback function should take one argument:
                    - stderr (str):
                        New incoming STDERR line string with trailing newline (backslash n).
            on_finish:
                Callback on terminating/completing a thread.
                If return value is not None then it is used for updating property `returnvalue`.
            poll_interval (float):
                Polling interval in seconds.
            quiet:
                No logging.
            subprocess_name:
                Subprocess name for logging.
        """
        super().__init__(
            target=self._popen,
            args=(args, cwd, stdin, on_poll, on_stdout, on_stderr, on_finish),
        )
        self._poll_interval = poll_interval
        self._quiet = quiet
        self._subprocess_name = subprocess_name

        self._stdout_list = []
        self._stderr_list = []
        self._returncode = None
        self._stop_it = False
        self._stop_signal = None
        self._status = None
        self._returnvalue = None

    @property
    def stdout(self):
        return ''.join(self._stdout_list)

    @property
    def stderr(self):
        return ''.join(self._stderr_list)

    @property
    def returncode(self):
        """Returns subprocess.Popen.returncode.
        None if not completed or any general Exception occurs.
        """
        return self._returncode

    @property
    def status(self):
        """Updated with return value of on_poll() for every polling.
        Also updated with return value of on_stdout() or on_stderr()
        if their return values are not None.
        """
        return self._status

    @property
    def returnvalue(self):
        """Updated with return value of on_finish()
        which is called when a thread is terminated.
        None if thread is still running so that on_finish() has not been called yet.
        This works like an actual return value of the function ran inside a thread.
        """
        return self._returnvalue

    def stop(self, stop_signal=DEFAULT_STOP_SIGNAL, wait=False):
        """Subprocess will be teminated after next polling.

        Args:
            wait:
                Wait for a valid returncode (which is not None).
        """
        self._stop_it = True
        self._stop_signal = stop_signal
        if wait:
            if self._returncode is None:
                logger.info(
                    '{name} stopped but waiting for graceful shutdown...'.format(
                        name=self._subprocess_name
                    )
                )
            while True:
                if self._returncode is not None:
                    return
                time.sleep(self._poll_interval)

    def _popen(
        self,
        args,
        cwd=None,
        stdin=None,
        on_poll=None,
        on_stdout=None,
        on_stderr=None,
        on_finish=None,
    ):
        """Wrapper for subprocess.Popen().
        """

        def read_stdout(stdout_bytes):
            text = stdout_bytes.decode()
            if text:
                self._stdout_list.append(text)
                if on_stdout:
                    ret_on_stdout = on_stdout(text)
                    if ret_on_stdout is not None:
                        self._status = ret_on_stdout

        def read_stderr(stderr_bytes):
            text = stderr_bytes.decode()
            if text:
                self._stderr_list.append(text)
                if on_stderr:
                    ret_on_stderr = on_stderr(text)
                    if ret_on_stderr is not None:
                        self._status = ret_on_stderr

        def read_from_stdout_obj(stdout):
            if is_fileobj_open(stdout):
                for line in iter(stdout.readline, b''):
                    read_stdout(line)

        def read_from_stderr_obj(stderr):
            if is_fileobj_open(stderr):
                for line in iter(stderr.readline, b''):
                    read_stderr(line)

        self._stop_it = False

        try:
            p = Popen(args, stdout=PIPE, stderr=PIPE, cwd=cwd, stdin=stdin)
            thread_stdout = Thread(
                target=read_from_stdout_obj, args=(p.stdout,), daemon=True
            )
            thread_stderr = Thread(
                target=read_from_stderr_obj, args=(p.stderr,), daemon=True
            )
            thread_stdout.start()
            thread_stderr.start()

            while True:
                if on_poll:
                    ret_on_poll = on_poll()
                    if ret_on_poll is not None:
                        self._status = ret_on_poll
                if p.poll() is not None:
                    self._returncode = p.poll()
                    break
                if self._stop_it and self._stop_signal:
                    p.send_signal(self._stop_signal)
                    break
                time.sleep(self._poll_interval)

        except Exception as e:
            if not self._quiet:
                logger.error(e, exc_info=True)

        finally:
            stdout_bytes, stderr_bytes = p.communicate()
            read_stdout(stdout_bytes)
            read_stderr(stderr_bytes)
            self._returncode = p.returncode

        if on_finish:
            ret_on_finish = on_finish()
            if ret_on_finish is not None:
                self._returnvalue = ret_on_finish

        if not self._quiet:
            if self._returncode:
                logger.error(
                    '{name} failed. returncode={rc}'.format(
                        name=self._subprocess_name, rc=self._returncode
                    )
                )
            else:
                logger.info(
                    '{name} finished successfully.'.format(name=self._subprocess_name)
                )



"""Variables and functions for backward_compatibililty
"""

CAPER_1_0_0_PARAM_KEY_NAME_CHANGE = {
    'out_dir': 'local_out_dir',
    'out_gcs_bucket': 'gcp_out_dir',
    'out_s3_bucket': 'aws_out_dir',
    'tmp_dir': 'local_loc_dir',
    'tmp_gcs_bucket': 'gcp_loc_dir',
    'tmp_s3_bucket': 'aws_loc_dir',
    'ip': 'hostname',
}

CAPER_1_4_2_PARAM_KEY_NAME_CHANGE = {'auto_update_metadata': 'auto_write_metadata'}

PARAM_KEY_NAME_CHANGE = {
    **CAPER_1_0_0_PARAM_KEY_NAME_CHANGE,
    **CAPER_1_4_2_PARAM_KEY_NAME_CHANGE,
}



import os
import re
from pathlib import Path

import setuptools

META_PATH = Path('caper', '__init__.py')
HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with Path(HERE, *parts).open(encoding='utf-8') as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise


with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='caper',
    version=find_meta('version'),
    python_requires='>=3.6',
    scripts=[
        'bin/caper',
        'bin/run_mysql_server_docker.sh',
        'bin/run_mysql_server_singularity.sh',
        'scripts/gcp_caper_server/create_instance.sh',
    ],
    author='Jin Lee',
    author_email='leepc12@gmail.com',
    description='Cromwell Assisted Pipeline ExecutoR',
    long_description='https://github.com/ENCODE-DCC/caper',
    long_description_content_type='text/markdown',
    url='https://github.com/ENCODE-DCC/caper',
    packages=setuptools.find_packages(exclude=['mysql*', 'docs']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
    ],
    install_requires=[
        'pyhocon>=0.3.53',
        'requests>=2.20',
        'pyopenssl',
        'autouri>=0.2.4',
        'miniwdl>=0.7.0',
        'humanfriendly',
        'numpy>=1.8.2',
        'pandas>=1.0',
        'scikit-learn>=0.19.2',
        'matplotlib>=1.5',
        'six>=1.13.0',
    ],
)



import logging

from .wdl_parser import WDLParser

logger = logging.getLogger(__name__)


class CaperWDLParser(WDLParser):
    """WDL parser for Caper.
    """

    RE_WDL_COMMENT_DOCKER = r'^\s*\#\s*CAPER\s+docker\s(.+)'
    RE_WDL_COMMENT_SINGULARITY = r'^\s*\#\s*CAPER\s+singularity\s(.+)'
    WDL_WORKFLOW_META_DOCKER = 'caper_docker'
    WDL_WORKFLOW_META_SINGULARITY = 'caper_singularity'

    def __init__(self, wdl):
        super().__init__(wdl)

    @property
    def caper_docker(self):
        """Find a Docker image in WDL for Caper.

        Backward compatibililty:
            Keep using old regex method
            if WDL_WORKFLOW_META_DOCKER doesn't exist in workflow's meta
        """
        if self.workflow_meta:
            if CaperWDLParser.WDL_WORKFLOW_META_DOCKER in self.workflow_meta:
                return self.workflow_meta[CaperWDLParser.WDL_WORKFLOW_META_DOCKER]

        ret = self._find_val_of_matched_lines(CaperWDLParser.RE_WDL_COMMENT_DOCKER)
        if ret:
            return ret[0].strip('"\'')

    @property
    def caper_singularity(self):
        """Find a Singularity image in WDL for Caper.

        Backward compatibililty:
            Keep using old regex method
            if WDL_WORKFLOW_META_SINGULARITY doesn't exist in workflow's meta
        """
        if self.workflow_meta:
            if CaperWDLParser.WDL_WORKFLOW_META_SINGULARITY in self.workflow_meta:
                return self.workflow_meta[CaperWDLParser.WDL_WORKFLOW_META_SINGULARITY]

        ret = self._find_val_of_matched_lines(CaperWDLParser.RE_WDL_COMMENT_SINGULARITY)
        if ret:
            return ret[0].strip('"\'')
